/*
* refit/scan/bootscreen.c
*
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are
* met:
*
*  * Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
*
*  * Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the
*    distribution.
*
*  * Neither the name of Christoph Pfisterer nor the names of the
*    contributors may be used to endorse or promote products derived
*    from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
* A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
* OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
* DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
* THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <Platform.h> // Only use angled for Platform, else, xcode project won't compile
#include "entry_scan.h"
#include "../refit/screen.h"
#include "../libeg/XImage.h"
#include "../libeg/XTheme.h"
#include "../Platform/Settings.h"
#include "../include/OSFlags.h"

extern "C" {
#include <Protocol/GraphicsOutput.h>
}

#ifndef DEBUG_ALL
#define DEBUG_BOOT_SCREEN 1
#else
#define DEBUG_BOOT_SCREEN DEBUG_ALL
#endif

#if DEBUG_BOOT_SCREEN == 0
#define DBG(...)
#else
#define DBG(...) DebugLog(DEBUG_BOOT_SCREEN, __VA_ARGS__)
#endif

STATIC const EFI_GRAPHICS_OUTPUT_BLT_PIXEL grayBackgroundPixel = { 0xBF, 0xBF, 0xBF, 0xFF };
STATIC UINT8 grayAppleLogo[] = {
   0x89, 0x50, 0x4E, 0x47, 0x0D, 0x0A, 0x1A, 0x0A, 0x00, 0x00, 0x00, 0x0D, 0x49, 0x48, 0x44, 0x52,
   0x00, 0x00, 0x00, 0x50, 0x00, 0x00, 0x00, 0x5F, 0x08, 0x06, 0x00, 0x00, 0x00, 0x7F, 0x47, 0x40,
   0x78, 0x00, 0x00, 0x0A, 0x30, 0x69, 0x43, 0x43, 0x50, 0x49, 0x43, 0x43, 0x20, 0x50, 0x72, 0x6F,
   0x66, 0x69, 0x6C, 0x65, 0x00, 0x00, 0x48, 0x89, 0x9D, 0x96, 0x77, 0x54, 0x54, 0xD7, 0x16, 0x87,
   0xCF, 0xBD, 0x77, 0x7A, 0xA1, 0xCD, 0x30, 0x14, 0x29, 0x43, 0xEF, 0xBD, 0x0D, 0x20, 0xBD, 0x37,
   0xA9, 0xD2, 0x44, 0x61, 0x98, 0x19, 0x60, 0x28, 0x03, 0x0E, 0x33, 0x34, 0xB1, 0x21, 0xA2, 0x02,
   0x11, 0x45, 0x44, 0x04, 0x15, 0x41, 0x82, 0x22, 0x06, 0x8C, 0x86, 0x22, 0xB1, 0x22, 0x8A, 0x85,
   0x80, 0x60, 0xC1, 0x1E, 0x90, 0x20, 0xA0, 0xC4, 0x60, 0x14, 0x51, 0x51, 0x79, 0x33, 0xB2, 0x56,
   0x74, 0xE5, 0xE5, 0xBD, 0x97, 0x97, 0xDF, 0x1F, 0x67, 0x7D, 0x6B, 0x9F, 0xBD, 0xF7, 0x3D, 0x67,
   0xEF, 0x7D, 0xD6, 0xBA, 0x00, 0x90, 0xBC, 0xFD, 0xB9, 0xBC, 0x74, 0x58, 0x0A, 0x80, 0x34, 0x9E,
   0x80, 0x1F, 0xE2, 0xE5, 0x4A, 0x8F, 0x8C, 0x8A, 0xA6, 0x63, 0xFB, 0x01, 0x0C, 0xF0, 0x00, 0x03,
   0xCC, 0x00, 0x60, 0xB2, 0x32, 0x33, 0x02, 0x42, 0x3D, 0xC3, 0x80, 0x48, 0x3E, 0x1E, 0x6E, 0xF4,
   0x4C, 0x91, 0x13, 0xF8, 0x22, 0x08, 0x80, 0x37, 0x77, 0xC4, 0x2B, 0x00, 0x37, 0x8D, 0xBC, 0x83,
   0xE8, 0x74, 0xF0, 0xFF, 0x49, 0x9A, 0x95, 0xC1, 0x17, 0x88, 0xD2, 0x04, 0x89, 0xD8, 0x82, 0xCD,
   0xC9, 0x64, 0x89, 0xB8, 0x50, 0xC4, 0xA9, 0xD9, 0x82, 0x0C, 0xB1, 0x7D, 0x46, 0xC4, 0xD4, 0xF8,
   0x14, 0x31, 0xC3, 0x28, 0x31, 0xF3, 0x45, 0x07, 0x14, 0xB1, 0xBC, 0x98, 0x13, 0x17, 0xD9, 0xF0,
   0xB3, 0xCF, 0x22, 0x3B, 0x8B, 0x99, 0x9D, 0xC6, 0x63, 0x8B, 0x58, 0x7C, 0xE6, 0x0C, 0x76, 0x1A,
   0x5B, 0xCC, 0x3D, 0x22, 0xDE, 0x9A, 0x25, 0xE4, 0x88, 0x18, 0xF1, 0x17, 0x71, 0x51, 0x16, 0x97,
   0x93, 0x2D, 0xE2, 0x5B, 0x22, 0xD6, 0x4C, 0x15, 0xA6, 0x71, 0x45, 0xFC, 0x56, 0x1C, 0x9B, 0xC6,
   0x61, 0x66, 0x02, 0x80, 0x22, 0x89, 0xED, 0x02, 0x0E, 0x2B, 0x49, 0xC4, 0xA6, 0x22, 0x26, 0xF1,
   0xC3, 0x42, 0xDC, 0x44, 0xBC, 0x14, 0x00, 0x1C, 0x29, 0xF1, 0x2B, 0x8E, 0xFF, 0x8A, 0x05, 0x9C,
   0x1C, 0x81, 0xF8, 0x52, 0x6E, 0xE9, 0x19, 0xB9, 0x7C, 0x6E, 0x62, 0x92, 0x80, 0xAE, 0xCB, 0xD2,
   0xA3, 0x9B, 0xD9, 0xDA, 0x32, 0xE8, 0xDE, 0x9C, 0xEC, 0x54, 0x8E, 0x40, 0x60, 0x14, 0xC4, 0x64,
   0xA5, 0x30, 0xF9, 0x6C, 0xBA, 0x5B, 0x7A, 0x5A, 0x06, 0x93, 0x97, 0x0B, 0xC0, 0xE2, 0x9D, 0x3F,
   0x4B, 0x46, 0x5C, 0x5B, 0xBA, 0xA8, 0xC8, 0xD6, 0x66, 0xB6, 0xD6, 0xD6, 0x46, 0xE6, 0xC6, 0x66,
   0x5F, 0x15, 0xEA, 0xBF, 0x6E, 0xFE, 0x4D, 0x89, 0x7B, 0xBB, 0x48, 0xAF, 0x82, 0x3F, 0xF7, 0x0C,
   0xA2, 0xF5, 0x7D, 0xB1, 0xFD, 0x95, 0x5F, 0x7A, 0x3D, 0x00, 0x8C, 0x59, 0x51, 0x6D, 0x76, 0x7C,
   0xB1, 0xC5, 0xEF, 0x05, 0xA0, 0x63, 0x33, 0x00, 0xF2, 0xF7, 0xBF, 0xD8, 0x34, 0x0F, 0x02, 0x20,
   0x29, 0xEA, 0x5B, 0xFB, 0xC0, 0x57, 0xF7, 0xA1, 0x89, 0xE7, 0x25, 0x49, 0x20, 0xC8, 0xB0, 0x33,
   0x31, 0xC9, 0xCE, 0xCE, 0x36, 0xE6, 0x72, 0x58, 0xC6, 0xE2, 0x82, 0xFE, 0xA1, 0xFF, 0xE9, 0xF0,
   0x37, 0xF4, 0xD5, 0xF7, 0x8C, 0xC5, 0xE9, 0xFE, 0x28, 0x0F, 0xDD, 0x9D, 0x93, 0xC0, 0x14, 0xA6,
   0x0A, 0xE8, 0xE2, 0xBA, 0xB1, 0xD2, 0x53, 0xD3, 0x85, 0x7C, 0x7A, 0x66, 0x06, 0x93, 0xC5, 0xA1,
   0x1B, 0xFD, 0x79, 0x88, 0xFF, 0x71, 0xE0, 0x5F, 0x9F, 0xC3, 0x30, 0x84, 0x93, 0xC0, 0xE1, 0x73,
   0x78, 0xA2, 0x88, 0x70, 0xD1, 0x94, 0x71, 0x79, 0x89, 0xA2, 0x76, 0xF3, 0xD8, 0x5C, 0x01, 0x37,
   0x9D, 0x47, 0xE7, 0xF2, 0xFE, 0x53, 0x13, 0xFF, 0x61, 0xD8, 0x9F, 0xB4, 0x38, 0xD7, 0x22, 0x51,
   0x1A, 0x3E, 0x01, 0x6A, 0xAC, 0x31, 0x90, 0x1A, 0xA0, 0x02, 0xE4, 0xD7, 0x3E, 0x80, 0xA2, 0x10,
   0x01, 0x12, 0x73, 0x40, 0xB4, 0x03, 0xFD, 0xD1, 0x37, 0x7F, 0x7C, 0x38, 0x10, 0xBF, 0xBC, 0x08,
   0xD5, 0x89, 0xC5, 0xB9, 0xFF, 0x2C, 0xE8, 0xDF, 0xB3, 0xC2, 0x65, 0xE2, 0x25, 0x93, 0x9B, 0xF8,
   0x39, 0xCE, 0x2D, 0x24, 0x8C, 0xCE, 0x12, 0xF2, 0xB3, 0x16, 0xF7, 0xC4, 0xCF, 0x12, 0xA0, 0x01,
   0x01, 0x48, 0x02, 0x2A, 0x50, 0x00, 0x2A, 0x40, 0x03, 0xE8, 0x02, 0x23, 0x60, 0x0E, 0x6C, 0x80,
   0x3D, 0x70, 0x06, 0x1E, 0xC0, 0x17, 0x04, 0x82, 0x30, 0x10, 0x05, 0x56, 0x01, 0x16, 0x48, 0x02,
   0x69, 0x80, 0x0F, 0xB2, 0x41, 0x3E, 0xD8, 0x08, 0x8A, 0x40, 0x09, 0xD8, 0x01, 0x76, 0x83, 0x6A,
   0x50, 0x0B, 0x1A, 0x40, 0x13, 0x68, 0x01, 0x27, 0x40, 0x07, 0x38, 0x0D, 0x2E, 0x80, 0xCB, 0xE0,
   0x3A, 0xB8, 0x01, 0x6E, 0x83, 0x07, 0x60, 0x04, 0x8C, 0x83, 0xE7, 0x60, 0x06, 0xBC, 0x01, 0xF3,
   0x10, 0x04, 0x61, 0x21, 0x32, 0x44, 0x81, 0x14, 0x20, 0x55, 0x48, 0x0B, 0x32, 0x80, 0xCC, 0x21,
   0x06, 0xE4, 0x08, 0x79, 0x40, 0xFE, 0x50, 0x08, 0x14, 0x05, 0xC5, 0x41, 0x89, 0x10, 0x0F, 0x12,
   0x42, 0xF9, 0xD0, 0x26, 0xA8, 0x04, 0x2A, 0x87, 0xAA, 0xA1, 0x3A, 0xA8, 0x09, 0xFA, 0x1E, 0x3A,
   0x05, 0x5D, 0x80, 0xAE, 0x42, 0x83, 0xD0, 0x3D, 0x68, 0x14, 0x9A, 0x82, 0x7E, 0x87, 0xDE, 0xC3,
   0x08, 0x4C, 0x82, 0xA9, 0xB0, 0x32, 0xAC, 0x0D, 0x9B, 0xC0, 0x0C, 0xD8, 0x05, 0xF6, 0x83, 0xC3,
   0xE0, 0x95, 0x70, 0x22, 0xBC, 0x1A, 0xCE, 0x83, 0x0B, 0xE1, 0xED, 0x70, 0x15, 0x5C, 0x0F, 0x1F,
   0x83, 0xDB, 0xE1, 0x0B, 0xF0, 0x75, 0xF8, 0x36, 0x3C, 0x02, 0x3F, 0x87, 0x67, 0x11, 0x80, 0x10,
   0x11, 0x1A, 0xA2, 0x86, 0x18, 0x21, 0x0C, 0xC4, 0x0D, 0x09, 0x44, 0xA2, 0x91, 0x04, 0x84, 0x8F,
   0xAC, 0x43, 0x8A, 0x91, 0x4A, 0xA4, 0x1E, 0x69, 0x41, 0xBA, 0x90, 0x5E, 0xE4, 0x26, 0x32, 0x82,
   0x4C, 0x23, 0xEF, 0x50, 0x18, 0x14, 0x05, 0x45, 0x47, 0x19, 0xA1, 0xEC, 0x51, 0xDE, 0xA8, 0xE5,
   0x28, 0x16, 0x6A, 0x35, 0x6A, 0x1D, 0xAA, 0x14, 0x55, 0x8D, 0x3A, 0x82, 0x6A, 0x47, 0xF5, 0xA0,
   0x6E, 0xA2, 0x46, 0x51, 0x33, 0xA8, 0x4F, 0x68, 0x32, 0x5A, 0x09, 0x6D, 0x80, 0xB6, 0x43, 0xFB,
   0xA0, 0x23, 0xD1, 0x89, 0xE8, 0x6C, 0x74, 0x11, 0xBA, 0x12, 0xDD, 0x88, 0x6E, 0x43, 0x5F, 0x42,
   0xDF, 0x46, 0x8F, 0xA3, 0xDF, 0x60, 0x30, 0x18, 0x1A, 0x46, 0x07, 0x63, 0x83, 0xF1, 0xC6, 0x44,
   0x61, 0x92, 0x31, 0x6B, 0x30, 0xA5, 0x98, 0xFD, 0x98, 0x56, 0xCC, 0x79, 0xCC, 0x20, 0x66, 0x0C,
   0x33, 0x8B, 0xC5, 0x62, 0x15, 0xB0, 0x06, 0x58, 0x07, 0x6C, 0x20, 0x96, 0x89, 0x15, 0x60, 0x8B,
   0xB0, 0x7B, 0xB1, 0xC7, 0xB0, 0xE7, 0xB0, 0x43, 0xD8, 0x71, 0xEC, 0x5B, 0x1C, 0x11, 0xA7, 0x8A,
   0x33, 0xC7, 0x79, 0xE2, 0xA2, 0x71, 0x3C, 0x5C, 0x01, 0xAE, 0x12, 0x77, 0x14, 0x77, 0x16, 0x37,
   0x84, 0x9B, 0xC0, 0xCD, 0xE3, 0xA5, 0xF0, 0x5A, 0x78, 0x3B, 0x7C, 0x20, 0x9E, 0x8D, 0xCF, 0xC5,
   0x97, 0xE1, 0x1B, 0xF0, 0x5D, 0xF8, 0x01, 0xFC, 0x38, 0x7E, 0x9E, 0x20, 0x4D, 0xD0, 0x21, 0x38,
   0x10, 0xC2, 0x08, 0xC9, 0x84, 0x8D, 0x84, 0x2A, 0x42, 0x0B, 0xE1, 0x12, 0xE1, 0x21, 0xE1, 0x15,
   0x91, 0x48, 0x54, 0x27, 0xDA, 0x12, 0x83, 0x89, 0x5C, 0xE2, 0x06, 0x62, 0x15, 0xF1, 0x38, 0xF1,
   0x0A, 0x71, 0x94, 0xF8, 0x8E, 0x24, 0x43, 0xD2, 0x27, 0xB9, 0x91, 0x62, 0x48, 0x42, 0xD2, 0x76,
   0xD2, 0x61, 0xD2, 0x79, 0xD2, 0x3D, 0xD2, 0x2B, 0x32, 0x99, 0xAC, 0x4D, 0x76, 0x26, 0x47, 0x93,
   0x05, 0xE4, 0xED, 0xE4, 0x26, 0xF2, 0x45, 0xF2, 0x63, 0xF2, 0x5B, 0x09, 0x8A, 0x84, 0xB1, 0x84,
   0x8F, 0x04, 0x5B, 0x62, 0xBD, 0x44, 0x8D, 0x44, 0xBB, 0xC4, 0x90, 0xC4, 0x0B, 0x49, 0xBC, 0xA4,
   0x96, 0xA4, 0x8B, 0xE4, 0x2A, 0xC9, 0x3C, 0xC9, 0x4A, 0xC9, 0x93, 0x92, 0x03, 0x92, 0xD3, 0x52,
   0x78, 0x29, 0x6D, 0x29, 0x37, 0x29, 0xA6, 0xD4, 0x3A, 0xA9, 0x1A, 0xA9, 0x53, 0x52, 0xC3, 0x52,
   0xB3, 0xD2, 0x14, 0x69, 0x33, 0xE9, 0x40, 0xE9, 0x34, 0xE9, 0x52, 0xE9, 0xA3, 0xD2, 0x57, 0xA5,
   0x27, 0x65, 0xB0, 0x32, 0xDA, 0x32, 0x1E, 0x32, 0x6C, 0x99, 0x42, 0x99, 0x43, 0x32, 0x17, 0x65,
   0xC6, 0x28, 0x08, 0x45, 0x83, 0xE2, 0x46, 0x61, 0x51, 0x36, 0x51, 0x1A, 0x28, 0x97, 0x28, 0xE3,
   0x54, 0x0C, 0x55, 0x87, 0xEA, 0x43, 0x4D, 0xA6, 0x96, 0x50, 0xBF, 0xA3, 0xF6, 0x53, 0x67, 0x64,
   0x65, 0x64, 0x2D, 0x65, 0xC3, 0x65, 0x73, 0x64, 0x6B, 0x64, 0xCF, 0xC8, 0x8E, 0xD0, 0x10, 0x9A,
   0x36, 0xCD, 0x87, 0x96, 0x4A, 0x2B, 0xA3, 0x9D, 0xA0, 0xDD, 0xA1, 0xBD, 0x97, 0x53, 0x96, 0x73,
   0x91, 0xE3, 0xC8, 0x6D, 0x93, 0x6B, 0x91, 0x1B, 0x92, 0x9B, 0x93, 0x5F, 0x22, 0xEF, 0x2C, 0xCF,
   0x91, 0x2F, 0x96, 0x6F, 0x95, 0xBF, 0x2D, 0xFF, 0x5E, 0x81, 0xAE, 0xE0, 0xA1, 0x90, 0xA2, 0xB0,
   0x53, 0xA1, 0x43, 0xE1, 0x91, 0x22, 0x4A, 0x51, 0x5F, 0x31, 0x58, 0x31, 0x5B, 0xF1, 0x80, 0xE2,
   0x25, 0xC5, 0xE9, 0x25, 0xD4, 0x25, 0xF6, 0x4B, 0x58, 0x4B, 0x8A, 0x97, 0x9C, 0x58, 0x72, 0x5F,
   0x09, 0x56, 0xD2, 0x57, 0x0A, 0x51, 0x5A, 0xA3, 0x74, 0x48, 0xA9, 0x4F, 0x69, 0x56, 0x59, 0x45,
   0xD9, 0x4B, 0x39, 0x43, 0x79, 0xAF, 0xF2, 0x45, 0xE5, 0x69, 0x15, 0x9A, 0x8A, 0xB3, 0x4A, 0xB2,
   0x4A, 0x85, 0xCA, 0x59, 0x95, 0x29, 0x55, 0x8A, 0xAA, 0xA3, 0x2A, 0x57, 0xB5, 0x42, 0xF5, 0x9C,
   0xEA, 0x33, 0xBA, 0x2C, 0xDD, 0x85, 0x9E, 0x4A, 0xAF, 0xA2, 0xF7, 0xD0, 0x67, 0xD4, 0x94, 0xD4,
   0xBC, 0xD5, 0x84, 0x6A, 0x75, 0x6A, 0xFD, 0x6A, 0xF3, 0xEA, 0x3A, 0xEA, 0xCB, 0xD5, 0x0B, 0xD4,
   0x5B, 0xD5, 0x1F, 0x69, 0x10, 0x34, 0x18, 0x1A, 0x09, 0x1A, 0x15, 0x1A, 0xDD, 0x1A, 0x33, 0x9A,
   0xAA, 0x9A, 0x01, 0x9A, 0xF9, 0x9A, 0xCD, 0x9A, 0xF7, 0xB5, 0xF0, 0x5A, 0x0C, 0xAD, 0x24, 0xAD,
   0x3D, 0x5A, 0xBD, 0x5A, 0x73, 0xDA, 0x3A, 0xDA, 0x11, 0xDA, 0x5B, 0xB4, 0x3B, 0xB4, 0x27, 0x75,
   0xE4, 0x75, 0x7C, 0x74, 0xF2, 0x74, 0x9A, 0x75, 0x1E, 0xEA, 0x92, 0x75, 0x9D, 0x74, 0x57, 0xEB,
   0xD6, 0xEB, 0xDE, 0xD2, 0xC3, 0xE8, 0x31, 0xF4, 0x52, 0xF4, 0xF6, 0xEB, 0xDD, 0xD0, 0x87, 0xF5,
   0xAD, 0xF4, 0x93, 0xF4, 0x6B, 0xF4, 0x07, 0x0C, 0x60, 0x03, 0x6B, 0x03, 0xAE, 0xC1, 0x7E, 0x83,
   0x41, 0x43, 0xB4, 0xA1, 0xAD, 0x21, 0xCF, 0xB0, 0xDE, 0x70, 0xD8, 0x88, 0x64, 0xE4, 0x62, 0x94,
   0x65, 0xD4, 0x6C, 0x34, 0x6A, 0x4C, 0x33, 0xF6, 0x37, 0x2E, 0x30, 0xEE, 0x30, 0x7E, 0x61, 0xA2,
   0x69, 0x12, 0x6D, 0xB2, 0xD3, 0xA4, 0xD7, 0xE4, 0x93, 0xA9, 0x95, 0x69, 0xAA, 0x69, 0x83, 0xE9,
   0x03, 0x33, 0x19, 0x33, 0x5F, 0xB3, 0x02, 0xB3, 0x2E, 0xB3, 0xDF, 0xCD, 0xF5, 0xCD, 0x59, 0xE6,
   0x35, 0xE6, 0xB7, 0x2C, 0xC8, 0x16, 0x9E, 0x16, 0xEB, 0x2D, 0x3A, 0x2D, 0x5E, 0x5A, 0x1A, 0x58,
   0x72, 0x2C, 0x0F, 0x58, 0xDE, 0xB5, 0xA2, 0x58, 0x05, 0x58, 0x6D, 0xB1, 0xEA, 0xB6, 0xFA, 0x68,
   0x6D, 0x63, 0xCD, 0xB7, 0x6E, 0xB1, 0x9E, 0xB2, 0xD1, 0xB4, 0x89, 0xB3, 0xD9, 0x67, 0x33, 0xCC,
   0xA0, 0x32, 0x82, 0x18, 0xA5, 0x8C, 0x2B, 0xB6, 0x68, 0x5B, 0x57, 0xDB, 0xF5, 0xB6, 0xA7, 0x6D,
   0xDF, 0xD9, 0x59, 0xDB, 0x09, 0xEC, 0x4E, 0xD8, 0xFD, 0x66, 0x6F, 0x64, 0x9F, 0x62, 0x7F, 0xD4,
   0x7E, 0x72, 0xA9, 0xCE, 0x52, 0xCE, 0xD2, 0x86, 0xA5, 0x63, 0x0E, 0xEA, 0x0E, 0x4C, 0x87, 0x3A,
   0x87, 0x11, 0x47, 0xBA, 0x63, 0x9C, 0xE3, 0x41, 0xC7, 0x11, 0x27, 0x35, 0x27, 0xA6, 0x53, 0xBD,
   0xD3, 0x13, 0x67, 0x0D, 0x67, 0xB6, 0x73, 0xA3, 0xF3, 0x84, 0x8B, 0x9E, 0x4B, 0xB2, 0xCB, 0x31,
   0x97, 0x17, 0xAE, 0xA6, 0xAE, 0x7C, 0xD7, 0x36, 0xD7, 0x39, 0x37, 0x3B, 0xB7, 0xB5, 0x6E, 0xE7,
   0xDD, 0x11, 0x77, 0x2F, 0xF7, 0x62, 0xF7, 0x7E, 0x0F, 0x19, 0x8F, 0xE5, 0x1E, 0xD5, 0x1E, 0x8F,
   0x3D, 0xD5, 0x3D, 0x13, 0x3D, 0x9B, 0x3D, 0x67, 0xBC, 0xAC, 0xBC, 0xD6, 0x78, 0x9D, 0xF7, 0x46,
   0x7B, 0xFB, 0x79, 0xEF, 0xF4, 0x1E, 0xF6, 0x51, 0xF6, 0x61, 0xF9, 0x34, 0xF9, 0xCC, 0xF8, 0xDA,
   0xF8, 0xAE, 0xF5, 0xED, 0xF1, 0x23, 0xF9, 0x85, 0xFA, 0x55, 0xFB, 0x3D, 0xF1, 0xD7, 0xF7, 0xE7,
   0xFB, 0x77, 0x05, 0xC0, 0x01, 0xBE, 0x01, 0xBB, 0x02, 0x1E, 0x2E, 0xD3, 0x5A, 0xC6, 0x5B, 0xD6,
   0x11, 0x08, 0x02, 0x7D, 0x02, 0x77, 0x05, 0x3E, 0x0A, 0xD2, 0x09, 0x5A, 0x1D, 0xF4, 0x63, 0x30,
   0x26, 0x38, 0x28, 0xB8, 0x26, 0xF8, 0x69, 0x88, 0x59, 0x48, 0x7E, 0x48, 0x6F, 0x28, 0x25, 0x34,
   0x36, 0xF4, 0x68, 0xE8, 0x9B, 0x30, 0xD7, 0xB0, 0xB2, 0xB0, 0x07, 0xCB, 0x75, 0x97, 0x0B, 0x97,
   0x77, 0x87, 0x4B, 0x86, 0xC7, 0x84, 0x37, 0x85, 0xCF, 0x45, 0xB8, 0x47, 0x94, 0x47, 0x8C, 0x44,
   0x9A, 0x44, 0xAE, 0x8D, 0xBC, 0x1E, 0xA5, 0x18, 0xC5, 0x8D, 0xEA, 0x8C, 0xC6, 0x46, 0x87, 0x47,
   0x37, 0x46, 0xCF, 0xAE, 0xF0, 0x58, 0xB1, 0x7B, 0xC5, 0x78, 0x8C, 0x55, 0x4C, 0x51, 0xCC, 0x9D,
   0x95, 0x3A, 0x2B, 0x73, 0x56, 0x5E, 0x5D, 0xA5, 0xB8, 0x2A, 0x75, 0xD5, 0x99, 0x58, 0xC9, 0x58,
   0x66, 0xEC, 0xC9, 0x38, 0x74, 0x5C, 0x44, 0xDC, 0xD1, 0xB8, 0x0F, 0xCC, 0x40, 0x66, 0x3D, 0x73,
   0x36, 0xDE, 0x27, 0x7E, 0x5F, 0xFC, 0x0C, 0xCB, 0x8D, 0xB5, 0x87, 0xF5, 0x9C, 0xED, 0xCC, 0xAE,
   0x60, 0x4F, 0x71, 0x1C, 0x38, 0xE5, 0x9C, 0x89, 0x04, 0x87, 0x84, 0xF2, 0x84, 0xC9, 0x44, 0x87,
   0xC4, 0x5D, 0x89, 0x53, 0x49, 0x4E, 0x49, 0x95, 0x49, 0xD3, 0x5C, 0x37, 0x6E, 0x35, 0xF7, 0x65,
   0xB2, 0x77, 0x72, 0x6D, 0xF2, 0x5C, 0x4A, 0x60, 0xCA, 0xE1, 0x94, 0x85, 0xD4, 0x88, 0xD4, 0xD6,
   0x34, 0x5C, 0x5A, 0x5C, 0xDA, 0x29, 0x9E, 0x0C, 0x2F, 0x85, 0xD7, 0x93, 0xAE, 0x92, 0x9E, 0x93,
   0x3E, 0x98, 0x61, 0x90, 0x51, 0x94, 0x31, 0xB2, 0xDA, 0x6E, 0xF5, 0xEE, 0xD5, 0x33, 0x7C, 0x3F,
   0x7E, 0x63, 0x26, 0x94, 0xB9, 0x32, 0xB3, 0x53, 0x40, 0x15, 0xFD, 0x4C, 0xF5, 0x09, 0x75, 0x85,
   0x9B, 0x85, 0xA3, 0x59, 0x8E, 0x59, 0x35, 0x59, 0x6F, 0xB3, 0xC3, 0xB3, 0x4F, 0xE6, 0x48, 0xE7,
   0xF0, 0x72, 0xFA, 0x72, 0xF5, 0x73, 0xB7, 0xE5, 0x4E, 0xE4, 0x79, 0xE6, 0x7D, 0xBB, 0x06, 0xB5,
   0x86, 0xB5, 0xA6, 0x3B, 0x5F, 0x2D, 0x7F, 0x63, 0xFE, 0xE8, 0x5A, 0x97, 0xB5, 0x75, 0xEB, 0xA0,
   0x75, 0xF1, 0xEB, 0xBA, 0xD7, 0x6B, 0xAC, 0x2F, 0x5C, 0x3F, 0xBE, 0xC1, 0x6B, 0xC3, 0x91, 0x8D,
   0x84, 0x8D, 0x29, 0x1B, 0x7F, 0x2A, 0x30, 0x2D, 0x28, 0x2F, 0x78, 0xBD, 0x29, 0x62, 0x53, 0x57,
   0xA1, 0x72, 0xE1, 0x86, 0xC2, 0xB1, 0xCD, 0x5E, 0x9B, 0x9B, 0x8B, 0x24, 0x8A, 0xF8, 0x45, 0xC3,
   0x5B, 0xEC, 0xB7, 0xD4, 0x6E, 0x45, 0x6D, 0xE5, 0x6E, 0xED, 0xDF, 0x66, 0xB1, 0x6D, 0xEF, 0xB6,
   0x4F, 0xC5, 0xEC, 0xE2, 0x6B, 0x25, 0xA6, 0x25, 0x95, 0x25, 0x1F, 0x4A, 0x59, 0xA5, 0xD7, 0xBE,
   0x31, 0xFB, 0xA6, 0xEA, 0x9B, 0x85, 0xED, 0x09, 0xDB, 0xFB, 0xCB, 0xAC, 0xCB, 0x0E, 0xEC, 0xC0,
   0xEC, 0xE0, 0xED, 0xB8, 0xB3, 0xD3, 0x69, 0xE7, 0x91, 0x72, 0xE9, 0xF2, 0xBC, 0xF2, 0xB1, 0x5D,
   0x01, 0xBB, 0xDA, 0x2B, 0xE8, 0x15, 0xC5, 0x15, 0xAF, 0x77, 0xC7, 0xEE, 0xBE, 0x5A, 0x69, 0x59,
   0x59, 0xBB, 0x87, 0xB0, 0x47, 0xB8, 0x67, 0xA4, 0xCA, 0xBF, 0xAA, 0x73, 0xAF, 0xE6, 0xDE, 0x1D,
   0x7B, 0x3F, 0x54, 0x27, 0x55, 0xDF, 0xAE, 0x71, 0xAD, 0x69, 0xDD, 0xA7, 0xB4, 0x6F, 0xDB, 0xBE,
   0xB9, 0xFD, 0xEC, 0xFD, 0x43, 0x07, 0x9C, 0x0F, 0xB4, 0xD4, 0x2A, 0xD7, 0x96, 0xD4, 0xBE, 0x3F,
   0xC8, 0x3D, 0x78, 0xB7, 0xCE, 0xAB, 0xAE, 0xBD, 0x5E, 0xBB, 0xBE, 0xF2, 0x10, 0xE6, 0x50, 0xD6,
   0xA1, 0xA7, 0x0D, 0xE1, 0x0D, 0xBD, 0xDF, 0x32, 0xBE, 0x6D, 0x6A, 0x54, 0x6C, 0x2C, 0x69, 0xFC,
   0x78, 0x98, 0x77, 0x78, 0xE4, 0x48, 0xC8, 0x91, 0x9E, 0x26, 0x9B, 0xA6, 0xA6, 0xA3, 0x4A, 0x47,
   0xCB, 0x9A, 0xE1, 0x66, 0x61, 0xF3, 0xD4, 0xB1, 0x98, 0x63, 0x37, 0xBE, 0x73, 0xFF, 0xAE, 0xB3,
   0xC5, 0xA8, 0xA5, 0xAE, 0x95, 0xD6, 0x5A, 0x72, 0x1C, 0x1C, 0x17, 0x1E, 0x7F, 0xF6, 0x7D, 0xDC,
   0xF7, 0x77, 0x4E, 0xF8, 0x9D, 0xE8, 0x3E, 0xC9, 0x38, 0xD9, 0xF2, 0x83, 0xD6, 0x0F, 0xFB, 0xDA,
   0x28, 0x6D, 0xC5, 0xED, 0x50, 0x7B, 0x6E, 0xFB, 0x4C, 0x47, 0x52, 0xC7, 0x48, 0x67, 0x54, 0xE7,
   0xE0, 0x29, 0xDF, 0x53, 0xDD, 0x5D, 0xF6, 0x5D, 0x6D, 0x3F, 0x1A, 0xFF, 0x78, 0xF8, 0xB4, 0xDA,
   0xE9, 0x9A, 0x33, 0xB2, 0x67, 0xCA, 0xCE, 0x12, 0xCE, 0x16, 0x9E, 0x5D, 0x38, 0x97, 0x77, 0x6E,
   0xF6, 0x7C, 0xC6, 0xF9, 0xE9, 0x0B, 0x89, 0x17, 0xC6, 0xBA, 0x63, 0xBB, 0x1F, 0x5C, 0x8C, 0xBC,
   0x78, 0xAB, 0x27, 0xB8, 0xA7, 0xFF, 0x92, 0xDF, 0xA5, 0x2B, 0x97, 0x3D, 0x2F, 0x5F, 0xEC, 0x75,
   0xE9, 0x3D, 0x77, 0xC5, 0xE1, 0xCA, 0xE9, 0xAB, 0x76, 0x57, 0x4F, 0x5D, 0x63, 0x5C, 0xEB, 0xB8,
   0x6E, 0x7D, 0xBD, 0xBD, 0xCF, 0xAA, 0xAF, 0xED, 0x27, 0xAB, 0x9F, 0xDA, 0xFA, 0xAD, 0xFB, 0xDB,
   0x07, 0x6C, 0x06, 0x3A, 0x6F, 0xD8, 0xDE, 0xE8, 0x1A, 0x5C, 0x3A, 0x78, 0x76, 0xC8, 0x69, 0xE8,
   0xC2, 0x4D, 0xF7, 0x9B, 0x97, 0x6F, 0xF9, 0xDC, 0xBA, 0x7E, 0x7B, 0xD9, 0xED, 0xC1, 0x3B, 0xCB,
   0xEF, 0xDC, 0x1D, 0x8E, 0x19, 0x1E, 0xB9, 0xCB, 0xBE, 0x3B, 0x79, 0x2F, 0xF5, 0xDE, 0xCB, 0xFB,
   0x59, 0xF7, 0xE7, 0x1F, 0x6C, 0x78, 0x88, 0x7E, 0x58, 0xFC, 0x48, 0xEA, 0x51, 0xE5, 0x63, 0xA5,
   0xC7, 0xF5, 0x3F, 0xEB, 0xFD, 0xDC, 0x3A, 0x62, 0x3D, 0x72, 0x66, 0xD4, 0x7D, 0xB4, 0xEF, 0x49,
   0xE8, 0x93, 0x07, 0x63, 0xAC, 0xB1, 0xE7, 0xBF, 0x64, 0xFE, 0xF2, 0x61, 0xBC, 0xF0, 0x29, 0xF9,
   0x69, 0xE5, 0x84, 0xEA, 0x44, 0xD3, 0xA4, 0xF9, 0xE4, 0xE9, 0x29, 0xCF, 0xA9, 0x1B, 0xCF, 0x56,
   0x3C, 0x1B, 0x7F, 0x9E, 0xF1, 0x7C, 0x7E, 0xBA, 0xE8, 0x57, 0xE9, 0x5F, 0xF7, 0xBD, 0xD0, 0x7D,
   0xF1, 0xC3, 0x6F, 0xCE, 0xBF, 0xF5, 0xCD, 0x44, 0xCE, 0x8C, 0xBF, 0xE4, 0xBF, 0x5C, 0xF8, 0xBD,
   0xF4, 0x95, 0xC2, 0xAB, 0xC3, 0xAF, 0x2D, 0x5F, 0x77, 0xCF, 0x06, 0xCD, 0x3E, 0x7E, 0x93, 0xF6,
   0x66, 0x7E, 0xAE, 0xF8, 0xAD, 0xC2, 0xDB, 0x23, 0xEF, 0x18, 0xEF, 0x7A, 0xDF, 0x47, 0xBC, 0x9F,
   0x98, 0xCF, 0xFE, 0x80, 0xFD, 0x50, 0xF5, 0x51, 0xEF, 0x63, 0xD7, 0x27, 0xBF, 0x4F, 0x0F, 0x17,
   0xD2, 0x16, 0x16, 0xFE, 0x05, 0x03, 0x98, 0xF3, 0xFC, 0xAA, 0xC7, 0x6A, 0xBE, 0x00, 0x00, 0x0C,
   0x43, 0x49, 0x44, 0x41, 0x54, 0x78, 0x01, 0xED, 0x9D, 0x7B, 0x4C, 0x54, 0xD9, 0x1D, 0xC7, 0x07,
   0x45, 0x14, 0x3B, 0x1A, 0xC1, 0x0A, 0xE3, 0x03, 0x8A, 0xEB, 0x68, 0x94, 0x6E, 0xC5, 0x28, 0x6E,
   0x7C, 0xA0, 0xA2, 0xC4, 0x14, 0x1F, 0xAB, 0xA2, 0x18, 0x35, 0x35, 0x82, 0xBA, 0xF1, 0x91, 0x68,
   0x6A, 0x34, 0x91, 0xA6, 0xDA, 0x44, 0x53, 0x36, 0xB1, 0x7F, 0xEE, 0xBF, 0xDD, 0x6D, 0x82, 0x49,
   0xDB, 0x65, 0xAD, 0x66, 0xBB, 0x66, 0xB7, 0xBB, 0x6D, 0x5C, 0x44, 0xC1, 0xD8, 0xCA, 0xB6, 0x05,
   0x63, 0xA8, 0x6D, 0x31, 0x35, 0x43, 0x13, 0x51, 0xE3, 0x34, 0x2A, 0x45, 0x1D, 0x5F, 0xFD, 0x7E,
   0x2F, 0xF7, 0x77, 0x73, 0xE6, 0x32, 0x8F, 0x3B, 0xC3, 0x0C, 0x1D, 0x38, 0x9E, 0xE4, 0xE7, 0xB9,
   0xDC, 0x7B, 0xEE, 0x79, 0x7C, 0xE6, 0xFB, 0x3B, 0xF7, 0xDC, 0x7B, 0xEE, 0x3D, 0xA6, 0xBD, 0x7E,
   0xFD, 0xDA, 0x15, 0x6F, 0x58, 0xB8, 0x70, 0x61, 0xBC, 0xA7, 0x26, 0xF5, 0xBC, 0xB4, 0xB4, 0x34,
   0x97, 0x62, 0x1E, 0xB4, 0x91, 0xD6, 0x65, 0x9A, 0x55, 0xF6, 0xD5, 0xAB, 0x57, 0xAD, 0xED, 0x78,
   0x37, 0xD2, 0xE3, 0x3D, 0x31, 0xD5, 0xCE, 0x53, 0x80, 0xB9, 0x86, 0x0D, 0x1B, 0x56, 0x02, 0x5B,
   0xCC, 0x18, 0xFB, 0xDD, 0x14, 0xC9, 0xA8, 0x51, 0xA3, 0x3E, 0x7E, 0xF4, 0xE8, 0xD1, 0xCF, 0x5F,
   0xBD, 0x7A, 0xE5, 0xE2, 0xDF, 0x89, 0x0A, 0x83, 0x1E, 0xA0, 0x80, 0x03, 0x2C, 0xF7, 0xF0, 0xE1,
   0xC3, 0x2B, 0x61, 0x9B, 0x60, 0xDC, 0x76, 0xD1, 0x18, 0x08, 0xAD, 0xA0, 0xA0, 0x20, 0xFB, 0xC6,
   0x8D, 0x1B, 0xC6, 0x76, 0xA2, 0xE0, 0x31, 0x9F, 0x41, 0x0D, 0x90, 0xF0, 0x00, 0xCE, 0x95, 0x9E,
   0x9E, 0x5E, 0x09, 0xAB, 0x86, 0x7D, 0x0B, 0x66, 0x80, 0xE3, 0x7E, 0x1A, 0xD5, 0xF6, 0xE2, 0xC5,
   0x0B, 0x63, 0x5B, 0x60, 0x6B, 0xAF, 0x40, 0x01, 0x01, 0x85, 0x79, 0x00, 0xEC, 0xA7, 0x23, 0x46,
   0x8C, 0xF0, 0xC2, 0x08, 0xD2, 0x80, 0x27, 0x60, 0xA9, 0x10, 0xC2, 0xA2, 0x02, 0xB9, 0x2F, 0x19,
   0x61, 0x58, 0x32, 0x32, 0x4D, 0x66, 0x9E, 0x02, 0x07, 0xB0, 0x4A, 0x46, 0x8E, 0x1C, 0xF9, 0x11,
   0xFA, 0x36, 0x2F, 0xCC, 0x85, 0x6D, 0x17, 0x21, 0x8A, 0xEB, 0x0A, 0x64, 0xA9, 0xCB, 0xAC, 0x59,
   0xB3, 0x3A, 0xB8, 0x9D, 0x48, 0xF5, 0x31, 0xBF, 0x41, 0xE5, 0xC2, 0x0A, 0xBC, 0xF2, 0x8C, 0x8C,
   0x8C, 0x1A, 0x98, 0x05, 0x8D, 0xEE, 0x1A, 0x4E, 0x65, 0x84, 0x06, 0xC0, 0x7F, 0x4F, 0x34, 0x3C,
   0x02, 0x1C, 0x34, 0x0A, 0x14, 0x78, 0x50, 0x59, 0x39, 0x60, 0xD4, 0x50, 0x71, 0x04, 0xA8, 0xBA,
   0x2D, 0x1B, 0x64, 0x0F, 0x84, 0x36, 0x7A, 0xF4, 0xE8, 0xEE, 0xDD, 0xBB, 0x77, 0xFF, 0x8B, 0xDB,
   0x89, 0x86, 0x38, 0x68, 0x14, 0x48, 0x80, 0x80, 0xC5, 0xBE, 0xCE, 0x80, 0x27, 0xEE, 0x1A, 0x4E,
   0x75, 0x04, 0x49, 0x58, 0xE6, 0x15, 0xF8, 0xD6, 0x8A, 0x15, 0x2B, 0xBA, 0x13, 0x0D, 0x8F, 0x65,
   0x0C, 0x0A, 0x05, 0x9A, 0x57, 0x5A, 0x37, 0x54, 0xF7, 0x3E, 0x95, 0xE7, 0x04, 0x9E, 0x0A, 0xB0,
   0xB8, 0xB8, 0xF8, 0x62, 0xA2, 0xC7, 0x7F, 0xCC, 0x9F, 0x21, 0xE5, 0x01, 0x8A, 0xEB, 0x66, 0x66,
   0x66, 0xEE, 0x82, 0x02, 0x73, 0x08, 0x2F, 0x52, 0x7F, 0xD7, 0xDB, 0xAC, 0xDE, 0x7F, 0xA9, 0xB8,
   0x71, 0xE3, 0xC6, 0xF9, 0x37, 0x6D, 0xDA, 0xD4, 0xA4, 0x35, 0x40, 0x0E, 0x57, 0x80, 0xA4, 0x42,
   0x94, 0x47, 0x80, 0xD1, 0x82, 0xB8, 0x2F, 0xD4, 0xF7, 0x0D, 0xDC, 0xD7, 0xAF, 0x25, 0x40, 0xAA,
   0x8F, 0xC3, 0x12, 0xB7, 0xDB, 0xFD, 0x9E, 0xC0, 0x8B, 0xD4, 0xE7, 0xA9, 0x50, 0x09, 0x2C, 0x37,
   0x37, 0xB7, 0xEB, 0xC0, 0x81, 0x03, 0xBF, 0x79, 0xF9, 0xF2, 0x65, 0xC2, 0x2F, 0x1E, 0x52, 0x56,
   0xF4, 0x9F, 0x52, 0x52, 0xFE, 0x1F, 0x62, 0x13, 0xA0, 0x07, 0x00, 0xCA, 0x78, 0xB5, 0x75, 0xEA,
   0xBA, 0x84, 0x47, 0x68, 0x1B, 0x37, 0x6E, 0xFC, 0x12, 0xEA, 0xEB, 0xD0, 0x12, 0xA0, 0xF4, 0x7D,
   0x50, 0xDF, 0xD6, 0x58, 0xE0, 0x89, 0xEB, 0x2E, 0x58, 0xB0, 0xA0, 0x6D, 0xE5, 0xCA, 0x95, 0x9F,
   0xF3, 0x36, 0x2E, 0x59, 0xEE, 0x4B, 0x4D, 0xA5, 0xEC, 0x30, 0x46, 0xDC, 0x17, 0xAA, 0x5B, 0x2A,
   0x30, 0x19, 0x47, 0x0A, 0x02, 0x2F, 0x27, 0x27, 0xA7, 0xAB, 0xA6, 0xA6, 0xA6, 0xAE, 0xB4, 0xB4,
   0xD4, 0x9F, 0x4C, 0xF5, 0xB1, 0x2E, 0x29, 0xEB, 0xC2, 0x26, 0x34, 0x0F, 0x14, 0x94, 0xE5, 0xC4,
   0x75, 0x09, 0x8F, 0x36, 0x76, 0xEC, 0x58, 0xFF, 0xF1, 0xE3, 0xC7, 0xEB, 0xE1, 0xBA, 0xAD, 0xC9,
   0x86, 0x97, 0xB2, 0x00, 0x09, 0x8F, 0x36, 0x71, 0xE2, 0xC4, 0xB7, 0x09, 0x2F, 0x1A, 0x40, 0x51,
   0xDE, 0x98, 0x31, 0x63, 0xFC, 0xB5, 0xB5, 0xB5, 0xF5, 0x93, 0x27, 0x4F, 0xFE, 0x92, 0xAE, 0x3B,
   0x10, 0x00, 0x53, 0xDA, 0x85, 0xD1, 0xF7, 0x7D, 0xD7, 0x29, 0xBC, 0xE9, 0xD3, 0xA7, 0x77, 0x1C,
   0x3C, 0x78, 0xF0, 0x0F, 0x93, 0x26, 0x4D, 0xFA, 0x6C, 0xE9, 0xD2, 0xA5, 0x81, 0x81, 0x80, 0x47,
   0x05, 0xA6, 0x34, 0x40, 0x28, 0x2A, 0xF3, 0xE1, 0xC3, 0x87, 0xAC, 0x67, 0x9F, 0x20, 0x2E, 0x0B,
   0xC0, 0x81, 0xAD, 0x5B, 0xB7, 0x5E, 0xA9, 0xAC, 0xAC, 0x6C, 0x40, 0xA2, 0x6B, 0x02, 0x8F, 0x17,
   0x8E, 0x81, 0x08, 0x29, 0x09, 0x50, 0x5C, 0x98, 0xB1, 0x1A, 0x08, 0x8D, 0x81, 0x31, 0xC6, 0x87,
   0x81, 0x79, 0xF3, 0xE6, 0xFD, 0x6D, 0xC7, 0x8E, 0x1D, 0xAD, 0x18, 0xEF, 0x35, 0x2C, 0x5F, 0xBE,
   0xDC, 0x27, 0x57, 0xDC, 0x81, 0x82, 0xC7, 0xBA, 0xA4, 0x24, 0x40, 0x81, 0x24, 0x7D, 0x9B, 0x0C,
   0x43, 0xE0, 0xD2, 0x01, 0xF4, 0x8B, 0x5D, 0x45, 0x45, 0x45, 0x9D, 0xAB, 0x56, 0xAD, 0xEA, 0x00,
   0xB8, 0x56, 0xA4, 0x6D, 0x5F, 0xB6, 0x6C, 0x99, 0xE1, 0xB2, 0x92, 0x8E, 0xE7, 0x0F, 0x54, 0x48,
   0x38, 0x40, 0x51, 0x4F, 0xA4, 0x06, 0x88, 0xFB, 0x85, 0x4B, 0x23, 0xC7, 0xD7, 0xAF, 0x5F, 0x7F,
   0x8D, 0xAA, 0x62, 0x9E, 0x50, 0x9B, 0x7F, 0xC2, 0x84, 0x09, 0xDD, 0x38, 0xA7, 0x0B, 0xE6, 0xA3,
   0x61, 0x98, 0xD2, 0x07, 0x5C, 0xB4, 0x3E, 0x93, 0x65, 0x4A, 0xFE, 0xDC, 0xEE, 0x6F, 0x48, 0x63,
   0x66, 0xF1, 0x06, 0x4E, 0x6B, 0x8A, 0x9B, 0x31, 0x36, 0x2B, 0x3F, 0x07, 0xDB, 0xB4, 0x5C, 0x18,
   0xEF, 0x61, 0xAD, 0x80, 0xB2, 0xF8, 0x48, 0xE9, 0x16, 0xAC, 0x15, 0x6A, 0x69, 0x65, 0xD9, 0x62,
   0x56, 0x22, 0x6C, 0x48, 0x5E, 0xCD, 0xCD, 0xCD, 0x6E, 0xFC, 0x49, 0x63, 0xF0, 0x63, 0x68, 0x12,
   0x60, 0x7A, 0x2A, 0x4D, 0xD4, 0xC6, 0xBF, 0x99, 0x1E, 0xE6, 0x46, 0xF9, 0x9C, 0x85, 0x2B, 0x82,
   0x79, 0x60, 0xAC, 0x43, 0xEF, 0x99, 0xE6, 0xBF, 0x66, 0x59, 0x2C, 0x97, 0x53, 0x9C, 0x46, 0x3D,
   0x2E, 0x5F, 0xBE, 0x6C, 0x3C, 0xA9, 0x0E, 0x4A, 0x18, 0xC3, 0x1F, 0xFD, 0x02, 0xB8, 0x78, 0xF1,
   0x62, 0x03, 0x1A, 0x2A, 0x5E, 0x0E, 0x33, 0xA6, 0x11, 0x09, 0x51, 0x4C, 0x1A, 0xC0, 0xD8, 0xAC,
   0xBC, 0xD5, 0x78, 0x00, 0xE8, 0xC6, 0x95, 0xB2, 0x19, 0xF1, 0x15, 0x58, 0xB3, 0x1C, 0x67, 0xCC,
   0xC0, 0x73, 0x68, 0xCC, 0x4B, 0x82, 0x1C, 0x63, 0x6C, 0x1E, 0x37, 0xA0, 0x21, 0x0D, 0x67, 0xE2,
   0xBC, 0x4C, 0xCB, 0x7B, 0x67, 0xC6, 0x72, 0xBE, 0x9C, 0xCB, 0x58, 0xCA, 0x20, 0x7C, 0x31, 0xD4,
   0xA1, 0x8B, 0xE5, 0x23, 0x3E, 0xD7, 0xD4, 0xD4, 0x44, 0x75, 0xC7, 0x14, 0xE2, 0x06, 0xB8, 0x64,
   0xC9, 0x92, 0xA0, 0x69, 0x44, 0xDE, 0x6E, 0xD9, 0x2B, 0xCF, 0x46, 0xA8, 0xC1, 0xDE, 0x00, 0x0E,
   0x35, 0x4C, 0x23, 0xCC, 0xAF, 0xD0, 0x90, 0x73, 0x30, 0xAA, 0xC3, 0x68, 0xAC, 0x7A, 0x2E, 0xB7,
   0x05, 0x0A, 0x00, 0xCD, 0x41, 0x59, 0xDF, 0x87, 0x95, 0xB3, 0x4C, 0x31, 0x82, 0x53, 0xE1, 0x85,
   0x2B, 0x5F, 0xF2, 0x27, 0x44, 0xA9, 0x83, 0x39, 0x6E, 0xFC, 0x0A, 0x7F, 0x9F, 0x86, 0x2A, 0x1D,
   0x83, 0x8C, 0x0B, 0x20, 0x3A, 0xED, 0x12, 0x3C, 0x1D, 0xF9, 0x11, 0xA0, 0xF5, 0x99, 0x46, 0x94,
   0x46, 0xDA, 0x2B, 0x2F, 0x30, 0xA4, 0xF2, 0x8C, 0xD9, 0x00, 0x69, 0x84, 0x0C, 0x7C, 0xD1, 0x00,
   0xDE, 0xFC, 0x9F, 0xC3, 0x71, 0xBA, 0xB9, 0xD1, 0x10, 0x33, 0x4F, 0x2A, 0x8C, 0xD0, 0x4A, 0x60,
   0x1E, 0x42, 0x53, 0x7F, 0x34, 0x15, 0x9C, 0xC0, 0x96, 0x32, 0xD5, 0x98, 0xE5, 0x32, 0x48, 0x3D,
   0xEC, 0x75, 0x78, 0xFE, 0xFC, 0x39, 0xA7, 0x41, 0x4F, 0x37, 0x36, 0x36, 0xD6, 0x19, 0x09, 0xA3,
   0xFC, 0x13, 0x13, 0x40, 0xAA, 0x0E, 0x95, 0xAE, 0x01, 0xBC, 0x12, 0x75, 0x3E, 0xC2, 0x5E, 0xF9,
   0x28, 0x65, 0x1A, 0x87, 0xA5, 0x21, 0x6C, 0x00, 0xB7, 0xA9, 0x04, 0x15, 0xA6, 0xD9, 0x30, 0xE3,
   0x75, 0x0C, 0x00, 0x24, 0x3C, 0x63, 0xB2, 0xDC, 0x0E, 0x2D, 0x9E, 0xB2, 0xD5, 0xFA, 0xD9, 0x41,
   0xF2, 0x87, 0x24, 0x44, 0x58, 0x07, 0xEC, 0x67, 0xD1, 0xFA, 0x48, 0xC7, 0x00, 0xA1, 0x3A, 0xCE,
   0x47, 0xD4, 0xC2, 0x72, 0x61, 0xD6, 0xAF, 0x2F, 0x8A, 0x53, 0x2B, 0x15, 0xEB, 0xB6, 0x34, 0x82,
   0xB1, 0x28, 0x42, 0xC0, 0x72, 0x1F, 0xCB, 0x20, 0x28, 0x31, 0xF9, 0x9B, 0xE5, 0x70, 0x3B, 0x11,
   0x41, 0xEA, 0xC0, 0x72, 0x15, 0x88, 0x3D, 0x81, 0x40, 0xE0, 0x87, 0x97, 0x2E, 0x5D, 0x0A, 0x7B,
   0xA1, 0x71, 0x04, 0x90, 0xF0, 0x30, 0x17, 0xF1, 0x01, 0xC0, 0x8D, 0x16, 0x78, 0xF2, 0xCB, 0x27,
   0xA2, 0xF2, 0x6A, 0x1E, 0xD2, 0x10, 0x89, 0xE5, 0x98, 0xFC, 0x50, 0x12, 0xCB, 0xFE, 0x44, 0xC7,
   0x2C, 0x97, 0x10, 0xE9, 0x11, 0x54, 0x22, 0x00, 0xF6, 0x3C, 0x7B, 0xF6, 0x2C, 0x2C, 0xC4, 0xA8,
   0x00, 0x05, 0x1E, 0x5C, 0xD6, 0x82, 0x97, 0xEC, 0x46, 0x24, 0x1A, 0x4A, 0x3C, 0xF9, 0x09, 0x44,
   0x00, 0x8C, 0x08, 0x31, 0x22, 0x40, 0xC0, 0xF3, 0x40, 0x79, 0xBF, 0x20, 0x3C, 0xF6, 0x79, 0xEC,
   0xB8, 0xA9, 0x3C, 0x5D, 0x82, 0x0A, 0x11, 0x6E, 0x7D, 0xBB, 0xA7, 0xA7, 0xE7, 0x20, 0xFA, 0x44,
   0x0E, 0xE6, 0xAD, 0x10, 0xF1, 0x4E, 0x04, 0xAF, 0x4C, 0xBC, 0x4F, 0xB7, 0xD5, 0x11, 0x1E, 0x09,
   0xD1, 0xD3, 0x28, 0x18, 0x0A, 0x07, 0x1C, 0xB2, 0xF1, 0x98, 0x2C, 0x0B, 0xBB, 0x9D, 0x01, 0xC4,
   0xA8, 0xBF, 0x1A, 0x00, 0xDF, 0x62, 0x9F, 0xA7, 0x9B, 0xF2, 0x2C, 0x79, 0x61, 0x83, 0x7D, 0xE2,
   0xCC, 0x99, 0x33, 0x3B, 0x0E, 0x1D, 0x3A, 0xD4, 0x88, 0xFB, 0xF0, 0xFF, 0xA8, 0xC7, 0xB8, 0x1D,
   0x52, 0x81, 0xA5, 0xA5, 0xA5, 0x1E, 0xC0, 0xAB, 0xD2, 0x19, 0x1E, 0xC1, 0xF1, 0x42, 0x82, 0xA7,
   0x3C, 0x2D, 0x98, 0xD9, 0x3B, 0x0F, 0x56, 0xD7, 0x60, 0x01, 0x42, 0x53, 0x43, 0x48, 0x80, 0x98,
   0xC4, 0xFE, 0x31, 0xC7, 0x5B, 0x34, 0x9D, 0xFA, 0x3C, 0x01, 0x23, 0xF0, 0xD6, 0xAE, 0x5D, 0x7B,
   0x65, 0xE7, 0xCE, 0x9D, 0xBF, 0xC6, 0xFE, 0x76, 0x39, 0x66, 0x8F, 0xFB, 0x5C, 0x11, 0x40, 0x9C,
   0x37, 0xE1, 0xDF, 0x13, 0x78, 0xEC, 0x07, 0x74, 0x0A, 0x36, 0x78, 0x75, 0x68, 0x7B, 0x58, 0x78,
   0xE4, 0xD2, 0x47, 0x81, 0x98, 0x94, 0xF9, 0x01, 0x33, 0xD1, 0xB1, 0xDF, 0x63, 0xBB, 0x79, 0xE5,
   0xE5, 0xF4, 0x00, 0x94, 0x47, 0x78, 0x61, 0x07, 0xD0, 0x84, 0xC7, 0x10, 0xA4, 0x40, 0xA8, 0xCF,
   0x83, 0xC1, 0x63, 0xB1, 0x8E, 0xF0, 0x08, 0x83, 0x00, 0x39, 0xAB, 0x77, 0xEA, 0xD4, 0xA9, 0xD3,
   0xF8, 0x33, 0x2A, 0x3C, 0x9E, 0x13, 0x04, 0x30, 0x2B, 0x2B, 0xAB, 0x4C, 0xE0, 0xE9, 0xEA, 0xBA,
   0xDB, 0xB7, 0x6F, 0xFF, 0x1D, 0xB8, 0x34, 0x13, 0x8E, 0x93, 0x10, 0x04, 0x10, 0xF2, 0x35, 0x1E,
   0x0F, 0xE9, 0x76, 0xE1, 0x50, 0x5D, 0x17, 0xC3, 0xB7, 0x4F, 0x9D, 0x80, 0x93, 0x34, 0x16, 0x40,
   0xB8, 0xAF, 0x1B, 0xEE, 0x3B, 0x85, 0xF0, 0x74, 0x55, 0xDF, 0x86, 0x0D, 0x1B, 0x3E, 0x03, 0x18,
   0xBF, 0xC0, 0x71, 0x12, 0x5B, 0x00, 0xE1, 0xFB, 0xC5, 0x84, 0xA7, 0x1B, 0x40, 0xAA, 0x8F, 0x36,
   0x63, 0xC6, 0x8C, 0x0E, 0x4C, 0x51, 0x34, 0x3A, 0x81, 0xA6, 0xA6, 0xB1, 0x00, 0xE2, 0x45, 0xC4,
   0x22, 0xDD, 0xE0, 0x09, 0x08, 0x0E, 0x98, 0xF3, 0xF2, 0xF2, 0x38, 0x50, 0x0E, 0xBA, 0x4D, 0x93,
   0xE3, 0x91, 0x62, 0x0B, 0xE0, 0x93, 0x27, 0x4F, 0xDE, 0xD6, 0xCD, 0x75, 0x09, 0x46, 0xFA, 0x3F,
   0xDC, 0x6D, 0xFC, 0x3E, 0x12, 0xA8, 0x70, 0xC7, 0x2C, 0x80, 0x50, 0xDF, 0x58, 0x1D, 0x15, 0xC8,
   0x71, 0x9F, 0xC7, 0xE3, 0x91, 0xA9, 0xD2, 0x70, 0x9C, 0xC2, 0xEE, 0xB7, 0x00, 0x3E, 0x7E, 0xFC,
   0x38, 0x47, 0x37, 0x05, 0x4A, 0xFF, 0x87, 0xD7, 0xE1, 0x6E, 0x85, 0x25, 0x14, 0xE5, 0x80, 0x05,
   0x90, 0xEA, 0x63, 0xD0, 0x11, 0x22, 0xDE, 0xBF, 0x89, 0xB9, 0xEF, 0x13, 0xAE, 0x06, 0xB5, 0xB2,
   0xB2, 0x32, 0x63, 0x02, 0x5C, 0x47, 0x78, 0x54, 0x21, 0x1F, 0x57, 0x09, 0x90, 0x58, 0x63, 0x03,
   0x60, 0x7E, 0x7E, 0xFE, 0x54, 0xDD, 0xE0, 0x09, 0x28, 0x02, 0xEC, 0xEC, 0xEC, 0x8C, 0x69, 0xEC,
   0x27, 0xE7, 0x32, 0x36, 0x00, 0xE2, 0x25, 0x9D, 0x4C, 0x75, 0xE7, 0x9B, 0x6D, 0xE7, 0x04, 0x7A,
   0x3B, 0x3E, 0xA4, 0xD7, 0x55, 0x81, 0x44, 0x75, 0xE7, 0xCE, 0x9D, 0xEF, 0x38, 0x47, 0x16, 0x9C,
   0xD2, 0x02, 0x18, 0xBC, 0x5B, 0x9F, 0xBF, 0x28, 0x1C, 0x3C, 0x7D, 0xFF, 0x76, 0xBC, 0x2D, 0x36,
   0x00, 0xE2, 0xF1, 0x4D, 0x3B, 0xFB, 0x02, 0x1D, 0x03, 0x01, 0x62, 0x08, 0x97, 0x17, 0x6F, 0xDB,
   0x45, 0x81, 0x8E, 0x5F, 0xA6, 0x89, 0xB7, 0xA0, 0x54, 0x3C, 0x8F, 0xF0, 0x68, 0x4F, 0x9F, 0x3E,
   0xED, 0x37, 0x40, 0xE3, 0x96, 0x26, 0x15, 0x1B, 0x99, 0xEC, 0x3A, 0x11, 0x20, 0xDE, 0x3C, 0x18,
   0x27, 0x43, 0xB9, 0x58, 0xCB, 0x13, 0x05, 0xF2, 0x93, 0x02, 0xE3, 0xB5, 0xB2, 0x58, 0x33, 0x18,
   0xCC, 0xE9, 0x45, 0x81, 0xBC, 0x89, 0xC0, 0xC3, 0x94, 0xB2, 0x78, 0xDA, 0x62, 0x01, 0xC4, 0xAF,
   0x60, 0xBC, 0xFD, 0xA9, 0x5B, 0x5F, 0x48, 0x88, 0x04, 0x88, 0x67, 0xA1, 0x15, 0xFD, 0x02, 0x88,
   0xC7, 0x39, 0x9D, 0xF1, 0x64, 0x30, 0xD8, 0xCF, 0x21, 0x40, 0x4E, 0x63, 0xE0, 0xA1, 0xC2, 0x78,
   0xB8, 0x71, 0x79, 0xAC, 0xED, 0xB1, 0x14, 0x88, 0x5F, 0xC1, 0x47, 0xF5, 0xE9, 0xAA, 0x40, 0x42,
   0xC4, 0x7C, 0xF8, 0x7B, 0x71, 0x03, 0xC4, 0x7B, 0x1F, 0xFF, 0xE4, 0xA3, 0x1D, 0x1D, 0x83, 0xA8,
   0x10, 0x6D, 0xA7, 0x0A, 0xAB, 0x63, 0x61, 0x60, 0x29, 0x70, 0xEF, 0xDE, 0xBD, 0x6D, 0x78, 0x95,
   0xC3, 0xE8, 0x07, 0x63, 0xC9, 0x60, 0x28, 0xA4, 0x15, 0x80, 0x7C, 0x99, 0x00, 0x0C, 0xAA, 0x30,
   0x3F, 0xE4, 0x75, 0xDA, 0x2E, 0x0B, 0x20, 0x4E, 0xF0, 0x63, 0x5A, 0xF3, 0x81, 0x6E, 0x2E, 0x2C,
   0xA0, 0x04, 0x22, 0xDF, 0x44, 0xE3, 0xE2, 0x16, 0x78, 0xB5, 0x4F, 0x3E, 0xAF, 0x90, 0x24, 0x21,
   0x63, 0x15, 0xA0, 0xAB, 0xA0, 0xA0, 0xC0, 0xB8, 0x23, 0xD1, 0x11, 0x22, 0x01, 0xF2, 0x6A, 0x6C,
   0xBE, 0x13, 0x94, 0x03, 0x90, 0x1F, 0xE0, 0xBB, 0xBB, 0xA8, 0x10, 0x83, 0x00, 0xA2, 0x1F, 0xBC,
   0xC6, 0x7E, 0x50, 0x47, 0x80, 0x94, 0x17, 0x01, 0xF2, 0x62, 0xC2, 0xB7, 0xD2, 0x00, 0xF0, 0x2D,
   0xBE, 0xD6, 0x1C, 0x0D, 0x62, 0x10, 0xC0, 0xEA, 0xEA, 0xEA, 0x6F, 0xB8, 0xCA, 0x8F, 0xAE, 0x00,
   0x09, 0x91, 0x4A, 0xA4, 0x0A, 0xE9, 0xCA, 0x26, 0xC4, 0x8F, 0xF8, 0x9A, 0x33, 0x8F, 0x85, 0x0A,
   0x41, 0x00, 0x91, 0xC0, 0x8F, 0x67, 0x83, 0xFF, 0x26, 0x40, 0x5D, 0x21, 0x12, 0xA0, 0x0A, 0x11,
   0x2A, 0xCC, 0x85, 0x7D, 0x88, 0x0B, 0x4B, 0xB5, 0x13, 0x80, 0xAE, 0x69, 0xD3, 0xA6, 0xFD, 0x45,
   0xD7, 0xE1, 0x8C, 0x00, 0xB2, 0x43, 0xC4, 0xE3, 0x2E, 0x5E, 0x58, 0xAA, 0x30, 0xC4, 0xF9, 0x18,
   0x6A, 0x0C, 0x1A, 0x6C, 0x87, 0x7A, 0xC9, 0x3C, 0x7F, 0xF3, 0xE6, 0xCD, 0x1F, 0x42, 0x81, 0x19,
   0xEC, 0x0F, 0x74, 0x0F, 0x14, 0x13, 0x4D, 0x3E, 0x7B, 0x30, 0x3F, 0x7D, 0x68, 0xC6, 0x97, 0x4C,
   0x3F, 0x21, 0x1B, 0xBB, 0x0B, 0x73, 0x9F, 0x0F, 0x93, 0x2C, 0xFF, 0xD0, 0xD9, 0x8D, 0x09, 0x41,
   0x82, 0xED, 0xC2, 0x62, 0xF4, 0x8D, 0xB8, 0xB0, 0xF4, 0x58, 0xC7, 0x65, 0x43, 0x8D, 0xB9, 0xDE,
   0x0A, 0x89, 0xEB, 0xDA, 0x0F, 0xAA, 0x2C, 0xB8, 0x4D, 0x97, 0x16, 0x90, 0x8C, 0xB1, 0x26, 0x8D,
   0xF5, 0xFA, 0x5B, 0x28, 0x05, 0xBA, 0x40, 0xF8, 0x0A, 0x67, 0xEB, 0xDF, 0x00, 0xB4, 0xA3, 0x74,
   0xB9, 0x0A, 0x0A, 0x0A, 0x7C, 0xAB, 0x57, 0xAF, 0xFE, 0xAB, 0x1C, 0x09, 0x09, 0x10, 0x07, 0xBB,
   0xE1, 0xC6, 0xC6, 0xC5, 0xE4, 0x0D, 0xC4, 0x5E, 0x54, 0xE4, 0x40, 0xAF, 0x2C, 0x2C, 0x2C, 0xFC,
   0x23, 0xF9, 0x44, 0x03, 0xE8, 0xC2, 0x77, 0x11, 0x9F, 0xE2, 0xE9, 0x44, 0xB7, 0xEE, 0x57, 0x64,
   0x01, 0x45, 0x0E, 0x1C, 0x23, 0xEF, 0xDB, 0xB7, 0xEF, 0x0B, 0xD9, 0xC7, 0x38, 0x9C, 0x02, 0x79,
   0xAC, 0x03, 0x9F, 0xB7, 0xFE, 0x89, 0x27, 0xEA, 0xAE, 0x42, 0xB6, 0x9F, 0x1C, 0xC8, 0x03, 0x5C,
   0x7C, 0x84, 0x23, 0x21, 0x12, 0x40, 0x17, 0x68, 0xD7, 0xBF, 0x51, 0x61, 0xEF, 0x42, 0xDE, 0xE4,
   0x40, 0x1E, 0x02, 0x4E, 0xE2, 0x88, 0x00, 0x91, 0x48, 0x7B, 0x15, 0xDA, 0xD4, 0xD7, 0xE7, 0x1D,
   0x9A, 0x68, 0x00, 0xB5, 0x57, 0x21, 0x5D, 0x37, 0x9C, 0xFA, 0xA8, 0xC2, 0xA8, 0x00, 0xA9, 0xC2,
   0x2D, 0x5B, 0xB6, 0x5C, 0xD0, 0xB1, 0x2F, 0x14, 0xF5, 0x61, 0x91, 0x9F, 0x26, 0x72, 0x20, 0x30,
   0x7B, 0x70, 0x02, 0xD0, 0xB5, 0x6E, 0xDD, 0xBA, 0x4F, 0x30, 0xE9, 0xE4, 0xD3, 0x09, 0xA2, 0xC0,
   0xE3, 0x5A, 0x84, 0xF8, 0x76, 0xA4, 0x4F, 0xDF, 0x27, 0x20, 0x1D, 0x01, 0x44, 0xE2, 0xAE, 0x23,
   0x47, 0x8E, 0xD4, 0x63, 0x14, 0xAE, 0xCD, 0x23, 0x7F, 0x02, 0x64, 0x7B, 0xF7, 0xEC, 0xD9, 0xF3,
   0x5B, 0xB4, 0xDF, 0x27, 0xC0, 0xEC, 0xB1, 0x53, 0x80, 0x2E, 0xBC, 0x43, 0xF8, 0x75, 0x45, 0x45,
   0x45, 0xA3, 0x0E, 0xB7, 0x78, 0xA2, 0xBE, 0x45, 0x8B, 0x16, 0xFD, 0x79, 0xEE, 0xDC, 0xB9, 0x41,
   0xE3, 0xBE, 0xB8, 0x01, 0xE2, 0xC4, 0xC0, 0xB6, 0x6D, 0xDB, 0x7E, 0xC5, 0x0F, 0xF1, 0x86, 0xB2,
   0x2B, 0x0B, 0x3C, 0xBA, 0x2E, 0x6E, 0x26, 0x7E, 0x89, 0x76, 0x5B, 0x77, 0x1D, 0x76, 0x78, 0xFC,
   0xDB, 0xB1, 0x02, 0xCD, 0x93, 0x7D, 0xFB, 0xF7, 0xEF, 0xFF, 0x04, 0xCF, 0xC7, 0x86, 0xEC, 0x53,
   0x6B, 0x71, 0x5D, 0x2E, 0x23, 0x8A, 0x36, 0x47, 0xFC, 0xD4, 0x35, 0x1E, 0x80, 0xBC, 0x99, 0xBE,
   0x8C, 0x7E, 0xE1, 0x0B, 0xBA, 0x32, 0x95, 0x38, 0x94, 0x02, 0xE1, 0xB1, 0x5D, 0xEC, 0xAA, 0xB8,
   0x8C, 0xA8, 0x93, 0xB6, 0xC5, 0xAA, 0x40, 0xE6, 0x19, 0xC0, 0xD3, 0x9A, 0x33, 0xFC, 0x9A, 0x7B,
   0x28, 0xB9, 0xB2, 0xC0, 0x9B, 0x3D, 0x7B, 0x76, 0x3B, 0xBB, 0x2A, 0xB6, 0x33, 0x59, 0x00, 0x99,
   0xAF, 0x7F, 0xD7, 0xAE, 0x5D, 0x75, 0x43, 0xA5, 0x3F, 0x54, 0xFB, 0xBD, 0x13, 0x27, 0x4E, 0xF0,
   0x5B, 0xE1, 0xB0, 0x57, 0x5D, 0x3B, 0xD4, 0x78, 0x14, 0x28, 0x79, 0x74, 0xF0, 0xC3, 0x64, 0x76,
   0xB6, 0x83, 0x59, 0x89, 0x84, 0x47, 0x93, 0xE5, 0x93, 0xD1, 0x38, 0x7E, 0x33, 0xE7, 0x38, 0xF4,
   0x07, 0xA0, 0x0B, 0x93, 0x2C, 0xCD, 0xC7, 0x8E, 0x1D, 0xAB, 0x67, 0xE1, 0x89, 0x84, 0x28, 0x8D,
   0x62, 0x9E, 0xA1, 0x8C, 0xC7, 0x13, 0x15, 0x98, 0x17, 0x9E, 0x38, 0x07, 0x6A, 0xCD, 0xE5, 0x93,
   0x63, 0xCD, 0x37, 0x3D, 0xD6, 0x13, 0xD4, 0xF4, 0x9C, 0x60, 0x61, 0x67, 0x8B, 0xC2, 0x33, 0x60,
   0x1B, 0xEF, 0xDD, 0xBB, 0xE7, 0xE1, 0x23, 0x6F, 0x3E, 0x02, 0x8F, 0x35, 0x08, 0x34, 0xC6, 0x98,
   0x5A, 0xED, 0x82, 0xF9, 0xC7, 0x8F, 0x1F, 0xFF, 0x5F, 0xAF, 0xD7, 0x6B, 0x7D, 0xC3, 0x81, 0xB7,
   0xE9, 0xDD, 0xF7, 0xEF, 0xDF, 0x77, 0xDF, 0xBE, 0x7D, 0x3B, 0xF7, 0xEE, 0xDD, 0xBB, 0x5C, 0xA5,
   0xD2, 0x28, 0xAB, 0x3F, 0xE5, 0xF1, 0xC7, 0x3F, 0x7C, 0xF8, 0xB0, 0xB1, 0x7C, 0x32, 0xEA, 0xEC,
   0xA8, 0xDF, 0x53, 0xDB, 0x16, 0x6A, 0x56, 0x4E, 0x3D, 0x1E, 0x71, 0x9B, 0x2B, 0x58, 0x72, 0xE6,
   0x0E, 0xCB, 0x21, 0x65, 0x20, 0xE1, 0x8A, 0x93, 0x27, 0x4F, 0xBE, 0x7B, 0xFD, 0xFA, 0xF5, 0x42,
   0xEE, 0x73, 0xD2, 0x28, 0x3B, 0xB4, 0xF9, 0xF3, 0xE7, 0xDF, 0x5A, 0xB3, 0x66, 0x8D, 0x0F, 0x6B,
   0xA5, 0xB2, 0x0F, 0x22, 0x38, 0x8E, 0xC1, 0x2C, 0x80, 0xD8, 0x96, 0x25, 0x41, 0xF9, 0x65, 0x95,
   0xA7, 0xAE, 0xAE, 0xCE, 0xDB, 0xD2, 0xD2, 0x32, 0x2D, 0x56, 0x98, 0x2C, 0x97, 0xCA, 0x9E, 0x32,
   0x65, 0x8A, 0xEF, 0xE8, 0xD1, 0xA3, 0x9F, 0x63, 0xED, 0x69, 0x5E, 0x71, 0x23, 0x8E, 0xF7, 0x70,
   0x3C, 0x64, 0xE8, 0x17, 0x40, 0xAE, 0xA1, 0x4A, 0xC5, 0x11, 0x18, 0x67, 0xF3, 0x1B, 0x1A, 0x1A,
   0xDE, 0x39, 0x7B, 0xF6, 0x6C, 0xF9, 0xF9, 0xF3, 0xE7, 0xE7, 0x63, 0x9D, 0x29, 0x2E, 0xCF, 0x69,
   0x14, 0xAA, 0xC2, 0x54, 0xA1, 0xE1, 0x9C, 0xC0, 0xD4, 0xA9, 0x53, 0x7D, 0x55, 0x55, 0x55, 0x6D,
   0xE6, 0xE7, 0x56, 0xBC, 0x61, 0xF7, 0xAD, 0xC0, 0x5A, 0xA9, 0xD2, 0x25, 0x30, 0xBD, 0x04, 0x51,
   0x1C, 0xF3, 0xA5, 0x5D, 0xB8, 0x70, 0x81, 0x6F, 0x0C, 0x78, 0x6F, 0xDE, 0xBC, 0xE9, 0x3D, 0x73,
   0xE6, 0xCC, 0xAC, 0xF6, 0xF6, 0x76, 0x2F, 0xD6, 0xB8, 0xCA, 0x90, 0x74, 0xE1, 0xCA, 0xE5, 0x93,
   0x65, 0xDC, 0xDF, 0xB7, 0x60, 0xED, 0x69, 0x4E, 0x0E, 0xD1, 0x62, 0x56, 0x9E, 0x55, 0x27, 0xB5,
   0x82, 0xB2, 0xD3, 0x69, 0x2C, 0xFF, 0xB7, 0x26, 0x2B, 0x2A, 0x10, 0x2F, 0x5E, 0xBC, 0x98, 0x8F,
   0xF3, 0xDF, 0x81, 0x3A, 0xE6, 0xB4, 0xB5, 0xB5, 0xE5, 0xC1, 0xED, 0xB8, 0x12, 0x88, 0xD1, 0x28,
   0x96, 0x85, 0x37, 0xC0, 0xB8, 0x1A, 0xAF, 0x1F, 0xAB, 0xF2, 0x76, 0xA2, 0x01, 0x06, 0x30, 0xA4,
   0x6F, 0x05, 0x34, 0x2E, 0x03, 0x6A, 0x28, 0x43, 0x85, 0x6C, 0xAF, 0x8B, 0x0A, 0x47, 0x7E, 0x3C,
   0xFC, 0x70, 0x54, 0x66, 0x21, 0xCC, 0x8B, 0x1F, 0xD0, 0x0B, 0x90, 0x74, 0xF1, 0xEC, 0x07, 0x0F,
   0x1E, 0x64, 0xB3, 0x6C, 0xE6, 0xC1, 0x49, 0x32, 0xB8, 0x6B, 0xB7, 0x59, 0x2E, 0x07, 0xC8, 0xBC,
   0x58, 0x50, 0xE9, 0xFD, 0x0A, 0xFD, 0x52, 0x60, 0x94, 0x92, 0xA9, 0x0E, 0xC2, 0xA4, 0xBB, 0x19,
   0x8D, 0x30, 0xD3, 0xD3, 0x25, 0x69, 0xAC, 0x7C, 0xBF, 0x1B, 0x60, 0xE6, 0xA9, 0x46, 0x52, 0x66,
   0x36, 0x76, 0xD2, 0xA4, 0xEC, 0x2E, 0x6C, 0xD3, 0x4D, 0x13, 0x5A, 0x6E, 0x32, 0x01, 0xA2, 0xAE,
   0x43, 0x3F, 0xFC, 0x0F, 0x3B, 0x64, 0xF7, 0x56, 0xBE, 0x7D, 0xE0, 0xF7, 0x00, 0x00, 0x00, 0x00,
   0x49, 0x45, 0x4E, 0x44, 0xAE, 0x42, 0x60, 0x82, 0x82,
};

STATIC const EFI_GRAPHICS_OUTPUT_BLT_PIXEL blackBackgroundPixel = { 0x00, 0x00, 0x00, 0xFF };
STATIC UINT8 whiteAppleLogo[] = {
   0x89, 0x50, 0x4E, 0x47, 0x0D, 0x0A, 0x1A, 0x0A, 0x00, 0x00, 0x00, 0x0D, 0x49, 0x48, 0x44, 0x52,
   0x00, 0x00, 0x00, 0x54, 0x00, 0x00, 0x00, 0x67, 0x08, 0x06, 0x00, 0x00, 0x00, 0x9E, 0x85, 0x65,
   0xC2, 0x00, 0x00, 0x0A, 0x30, 0x69, 0x43, 0x43, 0x50, 0x49, 0x43, 0x43, 0x20, 0x50, 0x72, 0x6F,
   0x66, 0x69, 0x6C, 0x65, 0x00, 0x00, 0x48, 0x89, 0x9D, 0x96, 0x77, 0x54, 0x54, 0xD7, 0x16, 0x87,
   0xCF, 0xBD, 0x77, 0x7A, 0xA1, 0xCD, 0x30, 0x14, 0x29, 0x43, 0xEF, 0xBD, 0x0D, 0x20, 0xBD, 0x37,
   0xA9, 0xD2, 0x44, 0x61, 0x98, 0x19, 0x60, 0x28, 0x03, 0x0E, 0x33, 0x34, 0xB1, 0x21, 0xA2, 0x02,
   0x11, 0x45, 0x44, 0x04, 0x15, 0x41, 0x82, 0x22, 0x06, 0x8C, 0x86, 0x22, 0xB1, 0x22, 0x8A, 0x85,
   0x80, 0x60, 0xC1, 0x1E, 0x90, 0x20, 0xA0, 0xC4, 0x60, 0x14, 0x51, 0x51, 0x79, 0x33, 0xB2, 0x56,
   0x74, 0xE5, 0xE5, 0xBD, 0x97, 0x97, 0xDF, 0x1F, 0x67, 0x7D, 0x6B, 0x9F, 0xBD, 0xF7, 0x3D, 0x67,
   0xEF, 0x7D, 0xD6, 0xBA, 0x00, 0x90, 0xBC, 0xFD, 0xB9, 0xBC, 0x74, 0x58, 0x0A, 0x80, 0x34, 0x9E,
   0x80, 0x1F, 0xE2, 0xE5, 0x4A, 0x8F, 0x8C, 0x8A, 0xA6, 0x63, 0xFB, 0x01, 0x0C, 0xF0, 0x00, 0x03,
   0xCC, 0x00, 0x60, 0xB2, 0x32, 0x33, 0x02, 0x42, 0x3D, 0xC3, 0x80, 0x48, 0x3E, 0x1E, 0x6E, 0xF4,
   0x4C, 0x91, 0x13, 0xF8, 0x22, 0x08, 0x80, 0x37, 0x77, 0xC4, 0x2B, 0x00, 0x37, 0x8D, 0xBC, 0x83,
   0xE8, 0x74, 0xF0, 0xFF, 0x49, 0x9A, 0x95, 0xC1, 0x17, 0x88, 0xD2, 0x04, 0x89, 0xD8, 0x82, 0xCD,
   0xC9, 0x64, 0x89, 0xB8, 0x50, 0xC4, 0xA9, 0xD9, 0x82, 0x0C, 0xB1, 0x7D, 0x46, 0xC4, 0xD4, 0xF8,
   0x14, 0x31, 0xC3, 0x28, 0x31, 0xF3, 0x45, 0x07, 0x14, 0xB1, 0xBC, 0x98, 0x13, 0x17, 0xD9, 0xF0,
   0xB3, 0xCF, 0x22, 0x3B, 0x8B, 0x99, 0x9D, 0xC6, 0x63, 0x8B, 0x58, 0x7C, 0xE6, 0x0C, 0x76, 0x1A,
   0x5B, 0xCC, 0x3D, 0x22, 0xDE, 0x9A, 0x25, 0xE4, 0x88, 0x18, 0xF1, 0x17, 0x71, 0x51, 0x16, 0x97,
   0x93, 0x2D, 0xE2, 0x5B, 0x22, 0xD6, 0x4C, 0x15, 0xA6, 0x71, 0x45, 0xFC, 0x56, 0x1C, 0x9B, 0xC6,
   0x61, 0x66, 0x02, 0x80, 0x22, 0x89, 0xED, 0x02, 0x0E, 0x2B, 0x49, 0xC4, 0xA6, 0x22, 0x26, 0xF1,
   0xC3, 0x42, 0xDC, 0x44, 0xBC, 0x14, 0x00, 0x1C, 0x29, 0xF1, 0x2B, 0x8E, 0xFF, 0x8A, 0x05, 0x9C,
   0x1C, 0x81, 0xF8, 0x52, 0x6E, 0xE9, 0x19, 0xB9, 0x7C, 0x6E, 0x62, 0x92, 0x80, 0xAE, 0xCB, 0xD2,
   0xA3, 0x9B, 0xD9, 0xDA, 0x32, 0xE8, 0xDE, 0x9C, 0xEC, 0x54, 0x8E, 0x40, 0x60, 0x14, 0xC4, 0x64,
   0xA5, 0x30, 0xF9, 0x6C, 0xBA, 0x5B, 0x7A, 0x5A, 0x06, 0x93, 0x97, 0x0B, 0xC0, 0xE2, 0x9D, 0x3F,
   0x4B, 0x46, 0x5C, 0x5B, 0xBA, 0xA8, 0xC8, 0xD6, 0x66, 0xB6, 0xD6, 0xD6, 0x46, 0xE6, 0xC6, 0x66,
   0x5F, 0x15, 0xEA, 0xBF, 0x6E, 0xFE, 0x4D, 0x89, 0x7B, 0xBB, 0x48, 0xAF, 0x82, 0x3F, 0xF7, 0x0C,
   0xA2, 0xF5, 0x7D, 0xB1, 0xFD, 0x95, 0x5F, 0x7A, 0x3D, 0x00, 0x8C, 0x59, 0x51, 0x6D, 0x76, 0x7C,
   0xB1, 0xC5, 0xEF, 0x05, 0xA0, 0x63, 0x33, 0x00, 0xF2, 0xF7, 0xBF, 0xD8, 0x34, 0x0F, 0x02, 0x20,
   0x29, 0xEA, 0x5B, 0xFB, 0xC0, 0x57, 0xF7, 0xA1, 0x89, 0xE7, 0x25, 0x49, 0x20, 0xC8, 0xB0, 0x33,
   0x31, 0xC9, 0xCE, 0xCE, 0x36, 0xE6, 0x72, 0x58, 0xC6, 0xE2, 0x82, 0xFE, 0xA1, 0xFF, 0xE9, 0xF0,
   0x37, 0xF4, 0xD5, 0xF7, 0x8C, 0xC5, 0xE9, 0xFE, 0x28, 0x0F, 0xDD, 0x9D, 0x93, 0xC0, 0x14, 0xA6,
   0x0A, 0xE8, 0xE2, 0xBA, 0xB1, 0xD2, 0x53, 0xD3, 0x85, 0x7C, 0x7A, 0x66, 0x06, 0x93, 0xC5, 0xA1,
   0x1B, 0xFD, 0x79, 0x88, 0xFF, 0x71, 0xE0, 0x5F, 0x9F, 0xC3, 0x30, 0x84, 0x93, 0xC0, 0xE1, 0x73,
   0x78, 0xA2, 0x88, 0x70, 0xD1, 0x94, 0x71, 0x79, 0x89, 0xA2, 0x76, 0xF3, 0xD8, 0x5C, 0x01, 0x37,
   0x9D, 0x47, 0xE7, 0xF2, 0xFE, 0x53, 0x13, 0xFF, 0x61, 0xD8, 0x9F, 0xB4, 0x38, 0xD7, 0x22, 0x51,
   0x1A, 0x3E, 0x01, 0x6A, 0xAC, 0x31, 0x90, 0x1A, 0xA0, 0x02, 0xE4, 0xD7, 0x3E, 0x80, 0xA2, 0x10,
   0x01, 0x12, 0x73, 0x40, 0xB4, 0x03, 0xFD, 0xD1, 0x37, 0x7F, 0x7C, 0x38, 0x10, 0xBF, 0xBC, 0x08,
   0xD5, 0x89, 0xC5, 0xB9, 0xFF, 0x2C, 0xE8, 0xDF, 0xB3, 0xC2, 0x65, 0xE2, 0x25, 0x93, 0x9B, 0xF8,
   0x39, 0xCE, 0x2D, 0x24, 0x8C, 0xCE, 0x12, 0xF2, 0xB3, 0x16, 0xF7, 0xC4, 0xCF, 0x12, 0xA0, 0x01,
   0x01, 0x48, 0x02, 0x2A, 0x50, 0x00, 0x2A, 0x40, 0x03, 0xE8, 0x02, 0x23, 0x60, 0x0E, 0x6C, 0x80,
   0x3D, 0x70, 0x06, 0x1E, 0xC0, 0x17, 0x04, 0x82, 0x30, 0x10, 0x05, 0x56, 0x01, 0x16, 0x48, 0x02,
   0x69, 0x80, 0x0F, 0xB2, 0x41, 0x3E, 0xD8, 0x08, 0x8A, 0x40, 0x09, 0xD8, 0x01, 0x76, 0x83, 0x6A,
   0x50, 0x0B, 0x1A, 0x40, 0x13, 0x68, 0x01, 0x27, 0x40, 0x07, 0x38, 0x0D, 0x2E, 0x80, 0xCB, 0xE0,
   0x3A, 0xB8, 0x01, 0x6E, 0x83, 0x07, 0x60, 0x04, 0x8C, 0x83, 0xE7, 0x60, 0x06, 0xBC, 0x01, 0xF3,
   0x10, 0x04, 0x61, 0x21, 0x32, 0x44, 0x81, 0x14, 0x20, 0x55, 0x48, 0x0B, 0x32, 0x80, 0xCC, 0x21,
   0x06, 0xE4, 0x08, 0x79, 0x40, 0xFE, 0x50, 0x08, 0x14, 0x05, 0xC5, 0x41, 0x89, 0x10, 0x0F, 0x12,
   0x42, 0xF9, 0xD0, 0x26, 0xA8, 0x04, 0x2A, 0x87, 0xAA, 0xA1, 0x3A, 0xA8, 0x09, 0xFA, 0x1E, 0x3A,
   0x05, 0x5D, 0x80, 0xAE, 0x42, 0x83, 0xD0, 0x3D, 0x68, 0x14, 0x9A, 0x82, 0x7E, 0x87, 0xDE, 0xC3,
   0x08, 0x4C, 0x82, 0xA9, 0xB0, 0x32, 0xAC, 0x0D, 0x9B, 0xC0, 0x0C, 0xD8, 0x05, 0xF6, 0x83, 0xC3,
   0xE0, 0x95, 0x70, 0x22, 0xBC, 0x1A, 0xCE, 0x83, 0x0B, 0xE1, 0xED, 0x70, 0x15, 0x5C, 0x0F, 0x1F,
   0x83, 0xDB, 0xE1, 0x0B, 0xF0, 0x75, 0xF8, 0x36, 0x3C, 0x02, 0x3F, 0x87, 0x67, 0x11, 0x80, 0x10,
   0x11, 0x1A, 0xA2, 0x86, 0x18, 0x21, 0x0C, 0xC4, 0x0D, 0x09, 0x44, 0xA2, 0x91, 0x04, 0x84, 0x8F,
   0xAC, 0x43, 0x8A, 0x91, 0x4A, 0xA4, 0x1E, 0x69, 0x41, 0xBA, 0x90, 0x5E, 0xE4, 0x26, 0x32, 0x82,
   0x4C, 0x23, 0xEF, 0x50, 0x18, 0x14, 0x05, 0x45, 0x47, 0x19, 0xA1, 0xEC, 0x51, 0xDE, 0xA8, 0xE5,
   0x28, 0x16, 0x6A, 0x35, 0x6A, 0x1D, 0xAA, 0x14, 0x55, 0x8D, 0x3A, 0x82, 0x6A, 0x47, 0xF5, 0xA0,
   0x6E, 0xA2, 0x46, 0x51, 0x33, 0xA8, 0x4F, 0x68, 0x32, 0x5A, 0x09, 0x6D, 0x80, 0xB6, 0x43, 0xFB,
   0xA0, 0x23, 0xD1, 0x89, 0xE8, 0x6C, 0x74, 0x11, 0xBA, 0x12, 0xDD, 0x88, 0x6E, 0x43, 0x5F, 0x42,
   0xDF, 0x46, 0x8F, 0xA3, 0xDF, 0x60, 0x30, 0x18, 0x1A, 0x46, 0x07, 0x63, 0x83, 0xF1, 0xC6, 0x44,
   0x61, 0x92, 0x31, 0x6B, 0x30, 0xA5, 0x98, 0xFD, 0x98, 0x56, 0xCC, 0x79, 0xCC, 0x20, 0x66, 0x0C,
   0x33, 0x8B, 0xC5, 0x62, 0x15, 0xB0, 0x06, 0x58, 0x07, 0x6C, 0x20, 0x96, 0x89, 0x15, 0x60, 0x8B,
   0xB0, 0x7B, 0xB1, 0xC7, 0xB0, 0xE7, 0xB0, 0x43, 0xD8, 0x71, 0xEC, 0x5B, 0x1C, 0x11, 0xA7, 0x8A,
   0x33, 0xC7, 0x79, 0xE2, 0xA2, 0x71, 0x3C, 0x5C, 0x01, 0xAE, 0x12, 0x77, 0x14, 0x77, 0x16, 0x37,
   0x84, 0x9B, 0xC0, 0xCD, 0xE3, 0xA5, 0xF0, 0x5A, 0x78, 0x3B, 0x7C, 0x20, 0x9E, 0x8D, 0xCF, 0xC5,
   0x97, 0xE1, 0x1B, 0xF0, 0x5D, 0xF8, 0x01, 0xFC, 0x38, 0x7E, 0x9E, 0x20, 0x4D, 0xD0, 0x21, 0x38,
   0x10, 0xC2, 0x08, 0xC9, 0x84, 0x8D, 0x84, 0x2A, 0x42, 0x0B, 0xE1, 0x12, 0xE1, 0x21, 0xE1, 0x15,
   0x91, 0x48, 0x54, 0x27, 0xDA, 0x12, 0x83, 0x89, 0x5C, 0xE2, 0x06, 0x62, 0x15, 0xF1, 0x38, 0xF1,
   0x0A, 0x71, 0x94, 0xF8, 0x8E, 0x24, 0x43, 0xD2, 0x27, 0xB9, 0x91, 0x62, 0x48, 0x42, 0xD2, 0x76,
   0xD2, 0x61, 0xD2, 0x79, 0xD2, 0x3D, 0xD2, 0x2B, 0x32, 0x99, 0xAC, 0x4D, 0x76, 0x26, 0x47, 0x93,
   0x05, 0xE4, 0xED, 0xE4, 0x26, 0xF2, 0x45, 0xF2, 0x63, 0xF2, 0x5B, 0x09, 0x8A, 0x84, 0xB1, 0x84,
   0x8F, 0x04, 0x5B, 0x62, 0xBD, 0x44, 0x8D, 0x44, 0xBB, 0xC4, 0x90, 0xC4, 0x0B, 0x49, 0xBC, 0xA4,
   0x96, 0xA4, 0x8B, 0xE4, 0x2A, 0xC9, 0x3C, 0xC9, 0x4A, 0xC9, 0x93, 0x92, 0x03, 0x92, 0xD3, 0x52,
   0x78, 0x29, 0x6D, 0x29, 0x37, 0x29, 0xA6, 0xD4, 0x3A, 0xA9, 0x1A, 0xA9, 0x53, 0x52, 0xC3, 0x52,
   0xB3, 0xD2, 0x14, 0x69, 0x33, 0xE9, 0x40, 0xE9, 0x34, 0xE9, 0x52, 0xE9, 0xA3, 0xD2, 0x57, 0xA5,
   0x27, 0x65, 0xB0, 0x32, 0xDA, 0x32, 0x1E, 0x32, 0x6C, 0x99, 0x42, 0x99, 0x43, 0x32, 0x17, 0x65,
   0xC6, 0x28, 0x08, 0x45, 0x83, 0xE2, 0x46, 0x61, 0x51, 0x36, 0x51, 0x1A, 0x28, 0x97, 0x28, 0xE3,
   0x54, 0x0C, 0x55, 0x87, 0xEA, 0x43, 0x4D, 0xA6, 0x96, 0x50, 0xBF, 0xA3, 0xF6, 0x53, 0x67, 0x64,
   0x65, 0x64, 0x2D, 0x65, 0xC3, 0x65, 0x73, 0x64, 0x6B, 0x64, 0xCF, 0xC8, 0x8E, 0xD0, 0x10, 0x9A,
   0x36, 0xCD, 0x87, 0x96, 0x4A, 0x2B, 0xA3, 0x9D, 0xA0, 0xDD, 0xA1, 0xBD, 0x97, 0x53, 0x96, 0x73,
   0x91, 0xE3, 0xC8, 0x6D, 0x93, 0x6B, 0x91, 0x1B, 0x92, 0x9B, 0x93, 0x5F, 0x22, 0xEF, 0x2C, 0xCF,
   0x91, 0x2F, 0x96, 0x6F, 0x95, 0xBF, 0x2D, 0xFF, 0x5E, 0x81, 0xAE, 0xE0, 0xA1, 0x90, 0xA2, 0xB0,
   0x53, 0xA1, 0x43, 0xE1, 0x91, 0x22, 0x4A, 0x51, 0x5F, 0x31, 0x58, 0x31, 0x5B, 0xF1, 0x80, 0xE2,
   0x25, 0xC5, 0xE9, 0x25, 0xD4, 0x25, 0xF6, 0x4B, 0x58, 0x4B, 0x8A, 0x97, 0x9C, 0x58, 0x72, 0x5F,
   0x09, 0x56, 0xD2, 0x57, 0x0A, 0x51, 0x5A, 0xA3, 0x74, 0x48, 0xA9, 0x4F, 0x69, 0x56, 0x59, 0x45,
   0xD9, 0x4B, 0x39, 0x43, 0x79, 0xAF, 0xF2, 0x45, 0xE5, 0x69, 0x15, 0x9A, 0x8A, 0xB3, 0x4A, 0xB2,
   0x4A, 0x85, 0xCA, 0x59, 0x95, 0x29, 0x55, 0x8A, 0xAA, 0xA3, 0x2A, 0x57, 0xB5, 0x42, 0xF5, 0x9C,
   0xEA, 0x33, 0xBA, 0x2C, 0xDD, 0x85, 0x9E, 0x4A, 0xAF, 0xA2, 0xF7, 0xD0, 0x67, 0xD4, 0x94, 0xD4,
   0xBC, 0xD5, 0x84, 0x6A, 0x75, 0x6A, 0xFD, 0x6A, 0xF3, 0xEA, 0x3A, 0xEA, 0xCB, 0xD5, 0x0B, 0xD4,
   0x5B, 0xD5, 0x1F, 0x69, 0x10, 0x34, 0x18, 0x1A, 0x09, 0x1A, 0x15, 0x1A, 0xDD, 0x1A, 0x33, 0x9A,
   0xAA, 0x9A, 0x01, 0x9A, 0xF9, 0x9A, 0xCD, 0x9A, 0xF7, 0xB5, 0xF0, 0x5A, 0x0C, 0xAD, 0x24, 0xAD,
   0x3D, 0x5A, 0xBD, 0x5A, 0x73, 0xDA, 0x3A, 0xDA, 0x11, 0xDA, 0x5B, 0xB4, 0x3B, 0xB4, 0x27, 0x75,
   0xE4, 0x75, 0x7C, 0x74, 0xF2, 0x74, 0x9A, 0x75, 0x1E, 0xEA, 0x92, 0x75, 0x9D, 0x74, 0x57, 0xEB,
   0xD6, 0xEB, 0xDE, 0xD2, 0xC3, 0xE8, 0x31, 0xF4, 0x52, 0xF4, 0xF6, 0xEB, 0xDD, 0xD0, 0x87, 0xF5,
   0xAD, 0xF4, 0x93, 0xF4, 0x6B, 0xF4, 0x07, 0x0C, 0x60, 0x03, 0x6B, 0x03, 0xAE, 0xC1, 0x7E, 0x83,
   0x41, 0x43, 0xB4, 0xA1, 0xAD, 0x21, 0xCF, 0xB0, 0xDE, 0x70, 0xD8, 0x88, 0x64, 0xE4, 0x62, 0x94,
   0x65, 0xD4, 0x6C, 0x34, 0x6A, 0x4C, 0x33, 0xF6, 0x37, 0x2E, 0x30, 0xEE, 0x30, 0x7E, 0x61, 0xA2,
   0x69, 0x12, 0x6D, 0xB2, 0xD3, 0xA4, 0xD7, 0xE4, 0x93, 0xA9, 0x95, 0x69, 0xAA, 0x69, 0x83, 0xE9,
   0x03, 0x33, 0x19, 0x33, 0x5F, 0xB3, 0x02, 0xB3, 0x2E, 0xB3, 0xDF, 0xCD, 0xF5, 0xCD, 0x59, 0xE6,
   0x35, 0xE6, 0xB7, 0x2C, 0xC8, 0x16, 0x9E, 0x16, 0xEB, 0x2D, 0x3A, 0x2D, 0x5E, 0x5A, 0x1A, 0x58,
   0x72, 0x2C, 0x0F, 0x58, 0xDE, 0xB5, 0xA2, 0x58, 0x05, 0x58, 0x6D, 0xB1, 0xEA, 0xB6, 0xFA, 0x68,
   0x6D, 0x63, 0xCD, 0xB7, 0x6E, 0xB1, 0x9E, 0xB2, 0xD1, 0xB4, 0x89, 0xB3, 0xD9, 0x67, 0x33, 0xCC,
   0xA0, 0x32, 0x82, 0x18, 0xA5, 0x8C, 0x2B, 0xB6, 0x68, 0x5B, 0x57, 0xDB, 0xF5, 0xB6, 0xA7, 0x6D,
   0xDF, 0xD9, 0x59, 0xDB, 0x09, 0xEC, 0x4E, 0xD8, 0xFD, 0x66, 0x6F, 0x64, 0x9F, 0x62, 0x7F, 0xD4,
   0x7E, 0x72, 0xA9, 0xCE, 0x52, 0xCE, 0xD2, 0x86, 0xA5, 0x63, 0x0E, 0xEA, 0x0E, 0x4C, 0x87, 0x3A,
   0x87, 0x11, 0x47, 0xBA, 0x63, 0x9C, 0xE3, 0x41, 0xC7, 0x11, 0x27, 0x35, 0x27, 0xA6, 0x53, 0xBD,
   0xD3, 0x13, 0x67, 0x0D, 0x67, 0xB6, 0x73, 0xA3, 0xF3, 0x84, 0x8B, 0x9E, 0x4B, 0xB2, 0xCB, 0x31,
   0x97, 0x17, 0xAE, 0xA6, 0xAE, 0x7C, 0xD7, 0x36, 0xD7, 0x39, 0x37, 0x3B, 0xB7, 0xB5, 0x6E, 0xE7,
   0xDD, 0x11, 0x77, 0x2F, 0xF7, 0x62, 0xF7, 0x7E, 0x0F, 0x19, 0x8F, 0xE5, 0x1E, 0xD5, 0x1E, 0x8F,
   0x3D, 0xD5, 0x3D, 0x13, 0x3D, 0x9B, 0x3D, 0x67, 0xBC, 0xAC, 0xBC, 0xD6, 0x78, 0x9D, 0xF7, 0x46,
   0x7B, 0xFB, 0x79, 0xEF, 0xF4, 0x1E, 0xF6, 0x51, 0xF6, 0x61, 0xF9, 0x34, 0xF9, 0xCC, 0xF8, 0xDA,
   0xF8, 0xAE, 0xF5, 0xED, 0xF1, 0x23, 0xF9, 0x85, 0xFA, 0x55, 0xFB, 0x3D, 0xF1, 0xD7, 0xF7, 0xE7,
   0xFB, 0x77, 0x05, 0xC0, 0x01, 0xBE, 0x01, 0xBB, 0x02, 0x1E, 0x2E, 0xD3, 0x5A, 0xC6, 0x5B, 0xD6,
   0x11, 0x08, 0x02, 0x7D, 0x02, 0x77, 0x05, 0x3E, 0x0A, 0xD2, 0x09, 0x5A, 0x1D, 0xF4, 0x63, 0x30,
   0x26, 0x38, 0x28, 0xB8, 0x26, 0xF8, 0x69, 0x88, 0x59, 0x48, 0x7E, 0x48, 0x6F, 0x28, 0x25, 0x34,
   0x36, 0xF4, 0x68, 0xE8, 0x9B, 0x30, 0xD7, 0xB0, 0xB2, 0xB0, 0x07, 0xCB, 0x75, 0x97, 0x0B, 0x97,
   0x77, 0x87, 0x4B, 0x86, 0xC7, 0x84, 0x37, 0x85, 0xCF, 0x45, 0xB8, 0x47, 0x94, 0x47, 0x8C, 0x44,
   0x9A, 0x44, 0xAE, 0x8D, 0xBC, 0x1E, 0xA5, 0x18, 0xC5, 0x8D, 0xEA, 0x8C, 0xC6, 0x46, 0x87, 0x47,
   0x37, 0x46, 0xCF, 0xAE, 0xF0, 0x58, 0xB1, 0x7B, 0xC5, 0x78, 0x8C, 0x55, 0x4C, 0x51, 0xCC, 0x9D,
   0x95, 0x3A, 0x2B, 0x73, 0x56, 0x5E, 0x5D, 0xA5, 0xB8, 0x2A, 0x75, 0xD5, 0x99, 0x58, 0xC9, 0x58,
   0x66, 0xEC, 0xC9, 0x38, 0x74, 0x5C, 0x44, 0xDC, 0xD1, 0xB8, 0x0F, 0xCC, 0x40, 0x66, 0x3D, 0x73,
   0x36, 0xDE, 0x27, 0x7E, 0x5F, 0xFC, 0x0C, 0xCB, 0x8D, 0xB5, 0x87, 0xF5, 0x9C, 0xED, 0xCC, 0xAE,
   0x60, 0x4F, 0x71, 0x1C, 0x38, 0xE5, 0x9C, 0x89, 0x04, 0x87, 0x84, 0xF2, 0x84, 0xC9, 0x44, 0x87,
   0xC4, 0x5D, 0x89, 0x53, 0x49, 0x4E, 0x49, 0x95, 0x49, 0xD3, 0x5C, 0x37, 0x6E, 0x35, 0xF7, 0x65,
   0xB2, 0x77, 0x72, 0x6D, 0xF2, 0x5C, 0x4A, 0x60, 0xCA, 0xE1, 0x94, 0x85, 0xD4, 0x88, 0xD4, 0xD6,
   0x34, 0x5C, 0x5A, 0x5C, 0xDA, 0x29, 0x9E, 0x0C, 0x2F, 0x85, 0xD7, 0x93, 0xAE, 0x92, 0x9E, 0x93,
   0x3E, 0x98, 0x61, 0x90, 0x51, 0x94, 0x31, 0xB2, 0xDA, 0x6E, 0xF5, 0xEE, 0xD5, 0x33, 0x7C, 0x3F,
   0x7E, 0x63, 0x26, 0x94, 0xB9, 0x32, 0xB3, 0x53, 0x40, 0x15, 0xFD, 0x4C, 0xF5, 0x09, 0x75, 0x85,
   0x9B, 0x85, 0xA3, 0x59, 0x8E, 0x59, 0x35, 0x59, 0x6F, 0xB3, 0xC3, 0xB3, 0x4F, 0xE6, 0x48, 0xE7,
   0xF0, 0x72, 0xFA, 0x72, 0xF5, 0x73, 0xB7, 0xE5, 0x4E, 0xE4, 0x79, 0xE6, 0x7D, 0xBB, 0x06, 0xB5,
   0x86, 0xB5, 0xA6, 0x3B, 0x5F, 0x2D, 0x7F, 0x63, 0xFE, 0xE8, 0x5A, 0x97, 0xB5, 0x75, 0xEB, 0xA0,
   0x75, 0xF1, 0xEB, 0xBA, 0xD7, 0x6B, 0xAC, 0x2F, 0x5C, 0x3F, 0xBE, 0xC1, 0x6B, 0xC3, 0x91, 0x8D,
   0x84, 0x8D, 0x29, 0x1B, 0x7F, 0x2A, 0x30, 0x2D, 0x28, 0x2F, 0x78, 0xBD, 0x29, 0x62, 0x53, 0x57,
   0xA1, 0x72, 0xE1, 0x86, 0xC2, 0xB1, 0xCD, 0x5E, 0x9B, 0x9B, 0x8B, 0x24, 0x8A, 0xF8, 0x45, 0xC3,
   0x5B, 0xEC, 0xB7, 0xD4, 0x6E, 0x45, 0x6D, 0xE5, 0x6E, 0xED, 0xDF, 0x66, 0xB1, 0x6D, 0xEF, 0xB6,
   0x4F, 0xC5, 0xEC, 0xE2, 0x6B, 0x25, 0xA6, 0x25, 0x95, 0x25, 0x1F, 0x4A, 0x59, 0xA5, 0xD7, 0xBE,
   0x31, 0xFB, 0xA6, 0xEA, 0x9B, 0x85, 0xED, 0x09, 0xDB, 0xFB, 0xCB, 0xAC, 0xCB, 0x0E, 0xEC, 0xC0,
   0xEC, 0xE0, 0xED, 0xB8, 0xB3, 0xD3, 0x69, 0xE7, 0x91, 0x72, 0xE9, 0xF2, 0xBC, 0xF2, 0xB1, 0x5D,
   0x01, 0xBB, 0xDA, 0x2B, 0xE8, 0x15, 0xC5, 0x15, 0xAF, 0x77, 0xC7, 0xEE, 0xBE, 0x5A, 0x69, 0x59,
   0x59, 0xBB, 0x87, 0xB0, 0x47, 0xB8, 0x67, 0xA4, 0xCA, 0xBF, 0xAA, 0x73, 0xAF, 0xE6, 0xDE, 0x1D,
   0x7B, 0x3F, 0x54, 0x27, 0x55, 0xDF, 0xAE, 0x71, 0xAD, 0x69, 0xDD, 0xA7, 0xB4, 0x6F, 0xDB, 0xBE,
   0xB9, 0xFD, 0xEC, 0xFD, 0x43, 0x07, 0x9C, 0x0F, 0xB4, 0xD4, 0x2A, 0xD7, 0x96, 0xD4, 0xBE, 0x3F,
   0xC8, 0x3D, 0x78, 0xB7, 0xCE, 0xAB, 0xAE, 0xBD, 0x5E, 0xBB, 0xBE, 0xF2, 0x10, 0xE6, 0x50, 0xD6,
   0xA1, 0xA7, 0x0D, 0xE1, 0x0D, 0xBD, 0xDF, 0x32, 0xBE, 0x6D, 0x6A, 0x54, 0x6C, 0x2C, 0x69, 0xFC,
   0x78, 0x98, 0x77, 0x78, 0xE4, 0x48, 0xC8, 0x91, 0x9E, 0x26, 0x9B, 0xA6, 0xA6, 0xA3, 0x4A, 0x47,
   0xCB, 0x9A, 0xE1, 0x66, 0x61, 0xF3, 0xD4, 0xB1, 0x98, 0x63, 0x37, 0xBE, 0x73, 0xFF, 0xAE, 0xB3,
   0xC5, 0xA8, 0xA5, 0xAE, 0x95, 0xD6, 0x5A, 0x72, 0x1C, 0x1C, 0x17, 0x1E, 0x7F, 0xF6, 0x7D, 0xDC,
   0xF7, 0x77, 0x4E, 0xF8, 0x9D, 0xE8, 0x3E, 0xC9, 0x38, 0xD9, 0xF2, 0x83, 0xD6, 0x0F, 0xFB, 0xDA,
   0x28, 0x6D, 0xC5, 0xED, 0x50, 0x7B, 0x6E, 0xFB, 0x4C, 0x47, 0x52, 0xC7, 0x48, 0x67, 0x54, 0xE7,
   0xE0, 0x29, 0xDF, 0x53, 0xDD, 0x5D, 0xF6, 0x5D, 0x6D, 0x3F, 0x1A, 0xFF, 0x78, 0xF8, 0xB4, 0xDA,
   0xE9, 0x9A, 0x33, 0xB2, 0x67, 0xCA, 0xCE, 0x12, 0xCE, 0x16, 0x9E, 0x5D, 0x38, 0x97, 0x77, 0x6E,
   0xF6, 0x7C, 0xC6, 0xF9, 0xE9, 0x0B, 0x89, 0x17, 0xC6, 0xBA, 0x63, 0xBB, 0x1F, 0x5C, 0x8C, 0xBC,
   0x78, 0xAB, 0x27, 0xB8, 0xA7, 0xFF, 0x92, 0xDF, 0xA5, 0x2B, 0x97, 0x3D, 0x2F, 0x5F, 0xEC, 0x75,
   0xE9, 0x3D, 0x77, 0xC5, 0xE1, 0xCA, 0xE9, 0xAB, 0x76, 0x57, 0x4F, 0x5D, 0x63, 0x5C, 0xEB, 0xB8,
   0x6E, 0x7D, 0xBD, 0xBD, 0xCF, 0xAA, 0xAF, 0xED, 0x27, 0xAB, 0x9F, 0xDA, 0xFA, 0xAD, 0xFB, 0xDB,
   0x07, 0x6C, 0x06, 0x3A, 0x6F, 0xD8, 0xDE, 0xE8, 0x1A, 0x5C, 0x3A, 0x78, 0x76, 0xC8, 0x69, 0xE8,
   0xC2, 0x4D, 0xF7, 0x9B, 0x97, 0x6F, 0xF9, 0xDC, 0xBA, 0x7E, 0x7B, 0xD9, 0xED, 0xC1, 0x3B, 0xCB,
   0xEF, 0xDC, 0x1D, 0x8E, 0x19, 0x1E, 0xB9, 0xCB, 0xBE, 0x3B, 0x79, 0x2F, 0xF5, 0xDE, 0xCB, 0xFB,
   0x59, 0xF7, 0xE7, 0x1F, 0x6C, 0x78, 0x88, 0x7E, 0x58, 0xFC, 0x48, 0xEA, 0x51, 0xE5, 0x63, 0xA5,
   0xC7, 0xF5, 0x3F, 0xEB, 0xFD, 0xDC, 0x3A, 0x62, 0x3D, 0x72, 0x66, 0xD4, 0x7D, 0xB4, 0xEF, 0x49,
   0xE8, 0x93, 0x07, 0x63, 0xAC, 0xB1, 0xE7, 0xBF, 0x64, 0xFE, 0xF2, 0x61, 0xBC, 0xF0, 0x29, 0xF9,
   0x69, 0xE5, 0x84, 0xEA, 0x44, 0xD3, 0xA4, 0xF9, 0xE4, 0xE9, 0x29, 0xCF, 0xA9, 0x1B, 0xCF, 0x56,
   0x3C, 0x1B, 0x7F, 0x9E, 0xF1, 0x7C, 0x7E, 0xBA, 0xE8, 0x57, 0xE9, 0x5F, 0xF7, 0xBD, 0xD0, 0x7D,
   0xF1, 0xC3, 0x6F, 0xCE, 0xBF, 0xF5, 0xCD, 0x44, 0xCE, 0x8C, 0xBF, 0xE4, 0xBF, 0x5C, 0xF8, 0xBD,
   0xF4, 0x95, 0xC2, 0xAB, 0xC3, 0xAF, 0x2D, 0x5F, 0x77, 0xCF, 0x06, 0xCD, 0x3E, 0x7E, 0x93, 0xF6,
   0x66, 0x7E, 0xAE, 0xF8, 0xAD, 0xC2, 0xDB, 0x23, 0xEF, 0x18, 0xEF, 0x7A, 0xDF, 0x47, 0xBC, 0x9F,
   0x98, 0xCF, 0xFE, 0x80, 0xFD, 0x50, 0xF5, 0x51, 0xEF, 0x63, 0xD7, 0x27, 0xBF, 0x4F, 0x0F, 0x17,
   0xD2, 0x16, 0x16, 0xFE, 0x05, 0x03, 0x98, 0xF3, 0xFC, 0xAA, 0xC7, 0x6A, 0xBE, 0x00, 0x00, 0x07,
   0xB7, 0x49, 0x44, 0x41, 0x54, 0x78, 0x01, 0xED, 0x9C, 0x7B, 0xA8, 0x15, 0x45, 0x1C, 0xC7, 0xBD,
   0xE6, 0xAB, 0x12, 0xD4, 0xD4, 0x84, 0xB0, 0xB8, 0x96, 0x3D, 0xFE, 0x30, 0x54, 0xD4, 0x42, 0x53,
   0x10, 0x13, 0xA4, 0xB2, 0xB2, 0x24, 0x25, 0xE8, 0x49, 0xD9, 0x8D, 0xFE, 0x29, 0x0A, 0x11, 0xAC,
   0x88, 0xB2, 0xFE, 0x31, 0x2A, 0x91, 0x0A, 0x8B, 0x08, 0x35, 0x23, 0x29, 0x7B, 0x18, 0x4A, 0x69,
   0x65, 0x0A, 0x95, 0x59, 0x56, 0x86, 0x99, 0x89, 0x95, 0x8F, 0x32, 0xB5, 0x32, 0x7C, 0x5D, 0xCB,
   0x47, 0xD9, 0xE7, 0x2B, 0x1E, 0xDB, 0xBB, 0x77, 0xCF, 0xEE, 0xCC, 0xEC, 0x9C, 0xDB, 0x9D, 0xB3,
   0xE7, 0x07, 0x1F, 0xEF, 0x3E, 0xE6, 0xF7, 0x9B, 0xF9, 0x7D, 0x9D, 0x9D, 0xB3, 0x3B, 0xB3, 0xE7,
   0xD4, 0x1D, 0x3D, 0x7A, 0xB4, 0x4D, 0xCD, 0xFC, 0x29, 0xD0, 0xD6, 0x5F, 0xA8, 0x5A, 0x24, 0x29,
   0x50, 0x13, 0xD4, 0x73, 0x3F, 0xA8, 0x09, 0x5A, 0x13, 0xD4, 0xB3, 0x02, 0x9E, 0xC3, 0x15, 0xA5,
   0x87, 0x2A, 0xCF, 0x8E, 0x9E, 0xB5, 0x4B, 0x0C, 0x57, 0xAD, 0x82, 0x9E, 0x4C, 0xB6, 0xE3, 0x61,
   0x0E, 0xFC, 0x08, 0x7F, 0xC3, 0x4E, 0xA8, 0xB8, 0xD5, 0x55, 0xD9, 0x6D, 0x53, 0x6F, 0x14, 0xBB,
   0x1B, 0x26, 0x41, 0x97, 0x04, 0xF5, 0xEA, 0x12, 0x8E, 0x79, 0x3D, 0xD4, 0xCE, 0x6B, 0xB4, 0xFF,
   0x2F, 0xD8, 0xA9, 0x54, 0x3D, 0x15, 0xEE, 0x85, 0x4E, 0x65, 0x9A, 0xF1, 0x57, 0x99, 0xE3, 0x5E,
   0x0F, 0x57, 0x83, 0xA0, 0x23, 0x50, 0x64, 0x2E, 0xD4, 0x67, 0x28, 0xF3, 0x47, 0xC6, 0x79, 0x2F,
   0xA7, 0x43, 0x1F, 0x43, 0x27, 0xA3, 0xC2, 0x72, 0xA8, 0x87, 0x2C, 0xDB, 0x98, 0x55, 0xC0, 0xC7,
   0xF9, 0x50, 0x7B, 0xA8, 0xC6, 0xC2, 0xA7, 0xE1, 0x2E, 0x0B, 0x11, 0xD6, 0x58, 0x94, 0x75, 0x2E,
   0x1A, 0x6A, 0x0F, 0x9D, 0x69, 0x29, 0xA6, 0x04, 0x5A, 0xAE, 0x7F, 0x2A, 0x6D, 0x21, 0x7E, 0xCA,
   0xAB, 0x57, 0x3E, 0x63, 0x29, 0xCC, 0x21, 0xCA, 0x9F, 0x06, 0x8D, 0x96, 0x7E, 0xD6, 0xC5, 0x43,
   0xEB, 0xA1, 0x03, 0xC8, 0x70, 0x86, 0x75, 0x96, 0x6D, 0xDA, 0xBC, 0x8E, 0x4F, 0xC5, 0xC5, 0x54,
   0xBB, 0x42, 0x12, 0x54, 0xE3, 0xFD, 0x6C, 0x68, 0x0F, 0xB6, 0xF6, 0xAC, 0xAD, 0x83, 0x6B, 0xF9,
   0x90, 0x04, 0x6D, 0x20, 0xC9, 0xFE, 0x0E, 0x89, 0xAE, 0xC0, 0xE7, 0x23, 0x07, 0x3F, 0x27, 0x97,
   0x50, 0xC6, 0x50, 0x3D, 0x4A, 0x6E, 0x81, 0x9E, 0x96, 0x59, 0x6A, 0xF6, 0x7C, 0x28, 0xAC, 0xB2,
   0xF4, 0x73, 0x2E, 0x1E, 0x4A, 0x0F, 0xBD, 0x81, 0x0C, 0x6D, 0xC5, 0x94, 0x28, 0xB3, 0xA0, 0xC5,
   0xC4, 0x54, 0x85, 0xA1, 0xF4, 0xD0, 0xD5, 0xB4, 0x75, 0x90, 0x1A, 0x6C, 0x61, 0x1B, 0x28, 0x3B,
   0x04, 0xF6, 0x59, 0xF8, 0xE4, 0x2E, 0x1A, 0x82, 0xA0, 0x7D, 0xC8, 0x52, 0x33, 0x46, 0x36, 0xB6,
   0x97, 0xC2, 0xC3, 0x61, 0xAD, 0x8D, 0x93, 0x8F, 0xB2, 0x21, 0x5C, 0xF2, 0x63, 0x2C, 0x13, 0xD5,
   0x24, 0xC8, 0x38, 0x68, 0x71, 0x31, 0xD5, 0xCE, 0x10, 0x04, 0xBD, 0x48, 0x0D, 0x35, 0xB4, 0x3D,
   0x94, 0xBB, 0x0C, 0x3E, 0x34, 0x2C, 0xEF, 0xBD, 0x58, 0x08, 0x82, 0x5E, 0x60, 0x98, 0xB5, 0x7A,
   0xA4, 0x3E, 0xD1, 0x97, 0x1B, 0x96, 0xAF, 0x48, 0xB1, 0x10, 0x04, 0xED, 0x91, 0x91, 0xF9, 0x61,
   0xCE, 0x4F, 0x07, 0xF5, 0xE4, 0xF5, 0x19, 0x65, 0x2B, 0x7E, 0x3A, 0x84, 0xD9, 0xA6, 0x0E, 0x65,
   0x54, 0xD0, 0xF3, 0xF9, 0x7C, 0x98, 0x06, 0xDF, 0x97, 0x29, 0xD3, 0xE2, 0x87, 0x43, 0x10, 0x54,
   0xE3, 0x62, 0xC9, 0xB6, 0xB3, 0xF1, 0x25, 0x2C, 0x06, 0x3D, 0x9F, 0xFF, 0x0A, 0xAD, 0xCA, 0x42,
   0xB8, 0x6D, 0x3A, 0x03, 0xC5, 0xF4, 0xFC, 0xFE, 0x1B, 0x1C, 0x68, 0x55, 0xEA, 0x25, 0x34, 0xA6,
   0x12, 0x82, 0x6A, 0x4D, 0x67, 0x30, 0x0C, 0x84, 0x7A, 0xE8, 0x05, 0x9D, 0x41, 0x62, 0xEC, 0x06,
   0x5D, 0x9E, 0xBA, 0x51, 0xFF, 0x1C, 0xFE, 0x04, 0x9F, 0xA6, 0x29, 0xBA, 0x91, 0xC7, 0xD1, 0xCC,
   0x94, 0xEE, 0x61, 0x75, 0x4C, 0x8F, 0xAE, 0xBA, 0x37, 0xDD, 0x05, 0x3B, 0x40, 0xB3, 0xF7, 0x5F,
   0x81, 0x9E, 0xF1, 0xD7, 0x80, 0x56, 0x45, 0xBD, 0x98, 0x2F, 0x41, 0xB5, 0x48, 0x36, 0x01, 0x26,
   0xC2, 0x48, 0x30, 0x59, 0x03, 0xD7, 0xFD, 0xE2, 0x52, 0x78, 0x13, 0x16, 0xC0, 0x7E, 0x70, 0xB1,
   0xD3, 0x71, 0x2A, 0xD5, 0x7D, 0x09, 0xDB, 0x75, 0x96, 0x41, 0xD4, 0xF3, 0x55, 0xFF, 0x4B, 0xB0,
   0xD2, 0xD2, 0xB7, 0x79, 0x71, 0x2D, 0x23, 0xE7, 0xA0, 0x17, 0xBE, 0xD3, 0x61, 0x2F, 0xE4, 0xB1,
   0x46, 0x9C, 0xE7, 0xC2, 0x28, 0x38, 0xF6, 0x9F, 0xCC, 0xDF, 0xB4, 0x76, 0xB5, 0xE7, 0xFC, 0x38,
   0x78, 0x1B, 0x8E, 0x80, 0x2F, 0x5B, 0x4D, 0xA0, 0xB1, 0x90, 0x56, 0x77, 0xEA, 0xB9, 0xD4, 0x93,
   0x29, 0x81, 0xDB, 0x71, 0x6E, 0x0A, 0xEC, 0x03, 0xDF, 0xB6, 0x89, 0x80, 0x0F, 0x41, 0x3F, 0x88,
   0xB6, 0xAF, 0x2D, 0xFB, 0xC3, 0xE0, 0x49, 0xD8, 0x09, 0x95, 0xB4, 0xF7, 0x08, 0x7E, 0x0E, 0x44,
   0xEB, 0x37, 0xDA, 0x76, 0xB9, 0xE4, 0xCF, 0xA3, 0x9F, 0xEB, 0x76, 0x45, 0x63, 0x64, 0xA5, 0x6D,
   0x27, 0x15, 0x68, 0xBC, 0x3B, 0x02, 0x17, 0x42, 0x77, 0x68, 0x29, 0x6B, 0xA4, 0xA2, 0x06, 0x78,
   0xD9, 0xA6, 0x42, 0x5B, 0x41, 0xAF, 0x24, 0xF8, 0x2B, 0xA0, 0x31, 0xB3, 0x28, 0xF6, 0x18, 0x89,
   0x3E, 0x08, 0x46, 0x6F, 0x26, 0xDB, 0x3C, 0x29, 0xDD, 0x41, 0xD0, 0xB7, 0xA0, 0x48, 0x62, 0x92,
   0x6E, 0x9B, 0xFB, 0x41, 0x4B, 0xD6, 0x46, 0x66, 0x2A, 0xE8, 0x6D, 0x44, 0x7B, 0x0E, 0x4C, 0xCB,
   0x1B, 0x55, 0x1E, 0x50, 0x21, 0xAD, 0xB4, 0x3E, 0x60, 0xD2, 0x5E, 0x93, 0x4B, 0x5E, 0xB3, 0x37,
   0x8B, 0xA0, 0xA8, 0x62, 0x46, 0x75, 0xBC, 0x9C, 0x9D, 0x77, 0xA2, 0x07, 0xE2, 0xDB, 0x59, 0x22,
   0xD5, 0xE3, 0xA0, 0x31, 0x33, 0xAB, 0x5C, 0x3C, 0x6E, 0x35, 0xEE, 0x1F, 0x26, 0xA9, 0xBE, 0x59,
   0x89, 0xA5, 0x3D, 0xCB, 0x4B, 0xC4, 0x79, 0x90, 0xF4, 0x5A, 0x60, 0x56, 0xDC, 0x6A, 0x3B, 0xBF,
   0x99, 0x84, 0xAE, 0x05, 0x3D, 0x5D, 0xA5, 0x5A, 0x9A, 0xA0, 0x77, 0xE2, 0xA9, 0x27, 0x8F, 0xA2,
   0xDB, 0x27, 0x08, 0x70, 0x0D, 0x18, 0x4D, 0xC4, 0x94, 0x1B, 0x43, 0xBB, 0x12, 0x40, 0xCF, 0xDC,
   0x2D, 0x79, 0xDF, 0x47, 0x75, 0xAD, 0xCE, 0x34, 0xF3, 0x3F, 0x16, 0x34, 0x0F, 0x61, 0x64, 0xE5,
   0xC6, 0xC6, 0x7B, 0xF0, 0x2E, 0xBA, 0x98, 0x5F, 0xA0, 0x81, 0xEE, 0xBB, 0x8D, 0xC5, 0x94, 0xE2,
   0x49, 0x3D, 0x54, 0x33, 0x33, 0x3F, 0x83, 0x66, 0x69, 0x8A, 0x6A, 0x7A, 0x42, 0x1B, 0x04, 0xDB,
   0x6C, 0x05, 0x48, 0xEA, 0xA1, 0xE3, 0x08, 0x52, 0x64, 0x31, 0xA5, 0xE1, 0xCD, 0x60, 0x2D, 0xA6,
   0x1C, 0x93, 0x04, 0x55, 0xB0, 0x22, 0xDB, 0x1C, 0x92, 0x5F, 0xE2, 0x2A, 0x40, 0xFC, 0x92, 0xEF,
   0x41, 0x20, 0x75, 0xF7, 0x24, 0xA1, 0x5D, 0xEB, 0x08, 0xC9, 0x4F, 0x13, 0x22, 0x7D, 0x41, 0x93,
   0xD0, 0x4E, 0x16, 0x17, 0x6E, 0x24, 0x51, 0xE2, 0xC7, 0x9C, 0x02, 0x07, 0xEA, 0xF4, 0x3C, 0xED,
   0x76, 0x16, 0x53, 0x39, 0xC7, 0xC5, 0xBB, 0x34, 0x50, 0x21, 0x7C, 0x35, 0x7B, 0x66, 0xDE, 0x40,
   0x71, 0x41, 0x87, 0xE7, 0x0D, 0x18, 0xB0, 0xFF, 0xC7, 0xB4, 0x7D, 0x73, 0xDE, 0xF6, 0x47, 0x05,
   0x3D, 0x89, 0x60, 0xE7, 0xE7, 0x0D, 0x18, 0xB0, 0xBF, 0x26, 0x80, 0x72, 0x5B, 0x54, 0x50, 0xAD,
   0x10, 0xB6, 0xCF, 0x1D, 0x31, 0xDC, 0x00, 0xCB, 0x7C, 0x34, 0x3D, 0x2A, 0xA8, 0x96, 0x36, 0x8A,
   0x6C, 0xEB, 0x7C, 0x24, 0x1F, 0x15, 0xB4, 0xC8, 0x37, 0xF3, 0xDB, 0x11, 0x53, 0xB7, 0x4C, 0xB9,
   0x2D, 0x2A, 0x68, 0xB7, 0xDC, 0xD1, 0xC2, 0x0D, 0xB0, 0xCB, 0x57, 0xD3, 0xA3, 0x82, 0x16, 0xB9,
   0x87, 0x1E, 0xAC, 0x84, 0xA0, 0xB6, 0x6F, 0x5C, 0xF8, 0x6A, 0x43, 0x6B, 0x88, 0xD3, 0xD1, 0x57,
   0x23, 0xA2, 0x3D, 0xD4, 0x57, 0xCC, 0x10, 0xE3, 0x68, 0xFE, 0xD7, 0x8B, 0x45, 0x05, 0xD5, 0xCB,
   0x04, 0x45, 0x35, 0xBD, 0x1F, 0xA5, 0xFB, 0xF0, 0xDC, 0x16, 0x15, 0xD4, 0xDB, 0xC0, 0x9C, 0xBB,
   0x55, 0x2D, 0x1F, 0xA0, 0x03, 0x55, 0x9E, 0xEB, 0xA3, 0xDA, 0xA8, 0xA0, 0xBF, 0xFB, 0x08, 0x18,
   0x70, 0x0C, 0x2F, 0xAF, 0x16, 0x45, 0x05, 0xD5, 0x6B, 0x7D, 0x45, 0xB6, 0x31, 0x3E, 0x92, 0x8F,
   0x0A, 0xBA, 0xC5, 0x47, 0xC0, 0x80, 0x63, 0x5C, 0x41, 0xDB, 0xD3, 0x56, 0x81, 0x8D, 0x52, 0x8B,
   0x0A, 0xBA, 0x15, 0x0F, 0x2D, 0xE6, 0x17, 0xD5, 0x34, 0xB9, 0xAE, 0xE5, 0x9F, 0x5C, 0x16, 0x15,
   0x54, 0xAF, 0x45, 0x6F, 0xCC, 0x15, 0x2D, 0x7C, 0x67, 0xAD, 0xF6, 0xE6, 0xB2, 0xA8, 0xA0, 0x0A,
   0xF4, 0x6D, 0xAE, 0x68, 0xE1, 0x3B, 0xEB, 0xC5, 0x0E, 0x2D, 0x1D, 0x3B, 0x5B, 0x5C, 0xD0, 0xCF,
   0x9C, 0x23, 0x55, 0x8F, 0xE3, 0x13, 0xA4, 0x72, 0x8A, 0x6B, 0x3A, 0x35, 0x41, 0x9B, 0x2B, 0xA7,
   0xFB, 0xD1, 0xC7, 0x9B, 0x1F, 0x36, 0x3B, 0x12, 0x5F, 0xF5, 0xEC, 0x8C, 0x9B, 0x7E, 0x81, 0xAB,
   0xC8, 0x13, 0xCD, 0x25, 0xE5, 0x6E, 0x65, 0x63, 0x76, 0x69, 0xC7, 0xF4, 0x6F, 0xBC, 0x87, 0xEE,
   0xC7, 0x71, 0xA5, 0xA9, 0x73, 0x95, 0x97, 0xD3, 0x0A, 0xE8, 0x55, 0xB6, 0x39, 0xC6, 0x05, 0x95,
   0xFF, 0xFB, 0xB6, 0x41, 0xAA, 0xB4, 0xBC, 0xAE, 0x52, 0x7D, 0x7F, 0x69, 0xBC, 0x4D, 0x7E, 0x49,
   0x82, 0xBE, 0x6B, 0x13, 0xA0, 0xCA, 0xCB, 0x4A, 0xD4, 0x57, 0x61, 0x2A, 0x18, 0x4D, 0x6F, 0xC6,
   0xC7, 0x50, 0xE9, 0x23, 0x47, 0x3D, 0x35, 0x9D, 0xA9, 0x9D, 0x9A, 0x9D, 0x50, 0x40, 0x8B, 0x78,
   0x93, 0x20, 0xF5, 0xE7, 0x3A, 0x92, 0x7A, 0xA8, 0xBE, 0x3E, 0xF2, 0xDA, 0x89, 0x30, 0xB5, 0x8D,
   0x92, 0x02, 0xA3, 0xD8, 0xF8, 0x06, 0xF4, 0x15, 0x1B, 0x7D, 0x78, 0x27, 0x5A, 0x52, 0x0F, 0x55,
   0xC1, 0x8B, 0xE1, 0xD3, 0x44, 0x8F, 0xDA, 0x41, 0x29, 0xA0, 0xA9, 0xCE, 0xDE, 0xA0, 0xEF, 0xAB,
   0x36, 0xB1, 0xA4, 0x1E, 0xAA, 0x02, 0xAB, 0xA0, 0xE8, 0x4F, 0x4D, 0x4D, 0x84, 0x8A, 0xED, 0x68,
   0x0D, 0x2A, 0x71, 0x1D, 0xAA, 0x9C, 0xA0, 0xF2, 0x7F, 0x31, 0x16, 0xA4, 0xB6, 0xFB, 0x9F, 0x02,
   0x73, 0xD8, 0xD4, 0xD0, 0xD8, 0xCC, 0xCA, 0x5D, 0xF2, 0x2A, 0xD8, 0x13, 0xB6, 0x42, 0x27, 0xED,
   0xD4, 0xAC, 0x89, 0x02, 0x7A, 0xE5, 0xF1, 0x87, 0x26, 0x47, 0x8E, 0xEF, 0xA4, 0xF5, 0x50, 0x4D,
   0x38, 0xCF, 0x4B, 0x72, 0x2A, 0xF8, 0x31, 0x7D, 0xDA, 0x27, 0x8A, 0x29, 0x5D, 0xD2, 0x04, 0xD5,
   0xF9, 0xA7, 0xF4, 0x4F, 0xCD, 0x9A, 0x28, 0x30, 0xA3, 0xC9, 0x5E, 0x6C, 0x27, 0x4B, 0x50, 0x7D,
   0x30, 0x2D, 0x8C, 0xF9, 0x14, 0x79, 0x77, 0x3D, 0xC9, 0x2F, 0x4E, 0x13, 0x20, 0x4B, 0x50, 0xF9,
   0x3E, 0x9C, 0x16, 0xA0, 0x60, 0xE7, 0x1E, 0x21, 0xDF, 0x7F, 0xD2, 0x72, 0x4E, 0xFB, 0x50, 0x8A,
   0xFA, 0xE9, 0xF1, 0xEB, 0xBA, 0xE8, 0x81, 0x02, 0x6E, 0xAB, 0x77, 0xF6, 0x03, 0x2F, 0x82, 0xF6,
   0x21, 0x90, 0x02, 0x7A, 0x7B, 0x65, 0x85, 0x58, 0xA1, 0xD9, 0x58, 0x1A, 0x9C, 0x7A, 0xB9, 0x2B,
   0x21, 0x93, 0x4B, 0x5E, 0xE5, 0x36, 0x41, 0x91, 0x3F, 0xA0, 0x34, 0x03, 0x97, 0x29, 0xA6, 0x84,
   0x32, 0xBD, 0xE4, 0x55, 0x56, 0xCF, 0xAF, 0x6B, 0xA1, 0x1E, 0x8A, 0x64, 0x87, 0x48, 0xB6, 0x3F,
   0x7C, 0x67, 0x92, 0xB4, 0x69, 0x0F, 0x55, 0x2C, 0x4D, 0x3E, 0xDF, 0x6E, 0x12, 0xB4, 0xCA, 0xCA,
   0x4C, 0x23, 0x1F, 0x23, 0x31, 0x95, 0xB7, 0x4D, 0x0F, 0x2D, 0xE9, 0x34, 0x8B, 0x8D, 0x86, 0xD2,
   0x4E, 0x95, 0xFF, 0xFD, 0x9A, 0xFC, 0x86, 0x80, 0xF1, 0xFB, 0x0A, 0x2E, 0x82, 0x6A, 0x45, 0x50,
   0x5F, 0xC4, 0xAF, 0xF6, 0x77, 0xF2, 0x35, 0xF9, 0xA1, 0x2F, 0xD0, 0xAE, 0x03, 0x63, 0xB3, 0xB9,
   0xE4, 0x4B, 0x41, 0x0F, 0xB0, 0x71, 0x3D, 0x68, 0x6C, 0xA9, 0x66, 0x9B, 0x42, 0x72, 0x56, 0x62,
   0x4A, 0x0C, 0x17, 0x41, 0xE5, 0xA7, 0x9F, 0x9C, 0xBC, 0x4F, 0x1B, 0x55, 0x6A, 0x8B, 0xC8, 0xCB,
   0xE9, 0x5B, 0x75, 0x2E, 0x97, 0x7C, 0x54, 0xC3, 0xF9, 0xEC, 0x4C, 0x8C, 0x1E, 0xA8, 0x82, 0xED,
   0xAD, 0xE4, 0xA0, 0x57, 0x1B, 0xB5, 0x9C, 0x6E, 0x6D, 0x79, 0x05, 0xD5, 0x78, 0xBA, 0x02, 0x06,
   0x5B, 0xD7, 0xDC, 0x3A, 0x1D, 0x34, 0x9C, 0x0D, 0x07, 0x7D, 0x46, 0x38, 0x99, 0xEB, 0x25, 0x5F,
   0xAA, 0x4C, 0x0D, 0xD0, 0xDA, 0xF5, 0x96, 0xD2, 0x81, 0x80, 0xFF, 0xEA, 0x91, 0xF2, 0x46, 0x70,
   0x16, 0x53, 0xB9, 0xE7, 0x15, 0x54, 0x31, 0xB6, 0xC3, 0x68, 0xD8, 0xA6, 0x9D, 0x40, 0x4D, 0x62,
   0xDE, 0x04, 0x6F, 0xE4, 0x6D, 0x7F, 0xDE, 0x4B, 0x3E, 0x5A, 0xBF, 0x66, 0xB1, 0x97, 0xC0, 0xD9,
   0xD1, 0x83, 0x8E, 0xDB, 0x6B, 0xF0, 0xD3, 0x1B, 0x2C, 0xEA, 0x2D, 0xBF, 0xC0, 0x6E, 0xD0, 0x5D,
   0x45, 0x37, 0xD0, 0xF7, 0xA9, 0x06, 0x80, 0x86, 0x99, 0x11, 0xA0, 0xF7, 0xE3, 0xF3, 0x98, 0xEE,
   0x31, 0x25, 0xA6, 0x3E, 0x0F, 0xF2, 0x9B, 0xCB, 0x6F, 0x64, 0xA6, 0xF8, 0x74, 0xE7, 0xDC, 0x52,
   0x70, 0xB1, 0x0D, 0x38, 0x4D, 0x86, 0xB3, 0xC0, 0xE8, 0xB7, 0x3A, 0x29, 0xD7, 0x05, 0x6E, 0x81,
   0x95, 0xE0, 0x62, 0x9B, 0x70, 0x1A, 0x0A, 0xA6, 0xF5, 0x65, 0x96, 0xCB, 0x2C, 0xE0, 0x50, 0x99,
   0x7A, 0x7D, 0x03, 0xEC, 0x80, 0x2C, 0x3B, 0x48, 0x81, 0x05, 0x30, 0x1A, 0x8E, 0x5D, 0x2D, 0xFC,
   0x75, 0x6D, 0xD3, 0x40, 0x7C, 0x5F, 0x80, 0x3D, 0x90, 0x65, 0xFA, 0x21, 0xD9, 0x47, 0xA1, 0x33,
   0xB8, 0xD6, 0x97, 0xE8, 0xE7, 0xF3, 0x92, 0x8F, 0x5F, 0x2E, 0xBA, 0x03, 0x98, 0x00, 0x57, 0x83,
   0x6E, 0x43, 0xBA, 0x82, 0xD6, 0xB1, 0x7F, 0x02, 0xDD, 0x30, 0x7F, 0x00, 0x8B, 0xC1, 0xE9, 0xF6,
   0x04, 0xBF, 0x72, 0xA6, 0x45, 0xC5, 0x31, 0x30, 0x0C, 0x34, 0x34, 0xA8, 0x5E, 0x1D, 0x6B, 0x04,
   0x3D, 0x93, 0x2F, 0x83, 0x85, 0xB0, 0x0F, 0xBC, 0x5B, 0x25, 0x05, 0xF5, 0xDE, 0xD8, 0x10, 0x02,
   0xFA, 0xF8, 0x94, 0x0F, 0x21, 0xCF, 0x16, 0x6B, 0x63, 0x4D, 0x50, 0xCF, 0x52, 0xD7, 0x04, 0xAD,
   0x09, 0xEA, 0x59, 0x01, 0xCF, 0xE1, 0xFE, 0x05, 0x40, 0x2F, 0xBA, 0x41, 0xFF, 0x13, 0x7D, 0xDC,
   0x00, 0x00, 0x00, 0x00, 0x49, 0x45, 0x4E, 0x44, 0xAE, 0x42, 0x60, 0x82, 0x82,
};

EFI_STATUS InitBootScreen(IN LOADER_ENTRY *Entry)
{
  EFI_GRAPHICS_OUTPUT_BLT_PIXEL thePixel = Entry->BootBgColor;
  XImage logo;
//  INTN      screenWidth, screenHeight;
  UINT8     customBoot = Entry->CustomBoot;

  if (OSFLAG_ISUNSET(Entry->Flags, OSFLAG_USEGRAPHICS)) {
    DBG("Custom boot screen not used because entry has unset use graphics\n");
    return EFI_ABORTED;
  }
  if (customBoot == CUSTOM_BOOT_USER) {
    logo = Entry->CustomLogo;
  } else if (customBoot == CUSTOM_BOOT_DISABLED) {
    customBoot = GlobalConfig.CustomLogoType;
    if (customBoot == CUSTOM_BOOT_USER) {
      logo = *GlobalConfig.CustomLogo;
    }
  }
  switch (customBoot) {
  case CUSTOM_BOOT_NONE:
     // No logo
     DBG("Custom boot is using no logo\n");
     break;

  case CUSTOM_BOOT_APPLE:
     // Gray on gray apple
     logo.FromPNG(grayAppleLogo, sizeof(grayAppleLogo));
     thePixel = grayBackgroundPixel;
     DBG("Custom boot is using apple logo\n");
     break;

  case CUSTOM_BOOT_ALT_APPLE:
     // Alternate white on black apple
     logo.FromPNG(whiteAppleLogo, sizeof(whiteAppleLogo));
     thePixel = blackBackgroundPixel;
     DBG("Custom boot is using alternate logo\n");
     break;

  case CUSTOM_BOOT_THEME:
     // TODO: Custom boot theme
      DBG("Custom boot is using theme logo\n");
      ThemeX.ClearScreen();
      return EFI_SUCCESS;

  case CUSTOM_BOOT_USER:
     // Custom user logo
     if (!logo.isEmpty()) {
       thePixel = grayBackgroundPixel;
       DBG("Custom boot is using custom logo\n");
       break;
     }

  default:
     DBG("Custom boot is disabled\n");
     return EFI_ABORTED;
  }

  // Clear the screen
  egClearScreen(&thePixel);
  // Draw the background logo
  if (!logo.isEmpty()) {
    logo.Draw((UGAWidth - logo.GetWidth()) / 2, (UGAHeight - logo.GetHeight()) / 2);
  }
  return EFI_SUCCESS;
}


static CONST CHAR8 *CustomBootModeStr[] = {
   "CUSTOM_BOOT_DISABLED",
   "CUSTOM_BOOT_USER_DISABLED",
   "CUSTOM_BOOT_NONE",
   "CUSTOM_BOOT_APPLE",
   "CUSTOM_BOOT_ALT_APPLE",
   "CUSTOM_BOOT_THEME",
   "CUSTOM_BOOT_USER",
};

CONST CHAR8 *CustomBootModeToStr(IN UINT8 Mode)
{
  if (Mode >= (sizeof(CustomBootModeStr) / sizeof(CustomBootModeStr[0]))) {
    return CustomBootModeStr[0];
  }
  return CustomBootModeStr[Mode];
}
